import os
import time
import pathlib

import click
import multiprocessing
from functools import partial

from chronix2grid.GeneratorBackend import GeneratorBackend
from chronix2grid import constants as cst
from chronix2grid.generation import generate_chronics as gen
from chronix2grid.generation import generation_utils as gu
from chronix2grid.kpi import main as kpis
from chronix2grid.output_processor import (
    output_processor_to_chunks, write_start_dates_for_chunks)
from chronix2grid.seed_manager import (parse_seed_arg, generate_default_seed,
                                       dump_seeds)
from chronix2grid import utils as ut


@click.command()
@click.option('--case', default='case118_l2rpn_neurips_1x', help='case folder to base generation on')
@click.option('--start-date', default='2012-01-01', help='Start date to generate chronics')
@click.option('--weeks', default=4, help='Number of weeks to generate')
@click.option('--by-n-weeks', default=4, help='Size of the output chunks in weeks')
@click.option('--n_scenarios', default=1, help='Number of scenarios to generate')
@click.option('--mode', default='LRTK', help='Steps to execute : '
                                              'L(K) for loads only (and KPI);R(K) for renewables (and KPI) only; '
                                              'LRT (K) for load, renewable and thermic generation (and KPI); '
                                              'LRDT(TK) for load, renewable, loss (dissipation) generation (and thermic and KPI)')
@click.option('--input-folder',
              default=os.path.join(pathlib.Path(__file__).parent.absolute(),
                                   cst.DEFAULT_INPUT_FOLDER_NAME),
              help='Directory to read input files from.')
@click.option('--output-folder',
              default=os.path.join(os.path.normpath(os.getcwd()),
                                   cst.DEFAULT_OUTPUT_FOLDER_NAME),
              help='Directory to store output files.')
@click.option('--seed-for-loads', default=None, help='Input seed to ensure reproducibility of loads generation')
@click.option('--seed-for-res', default=None, help='Input seed to ensure reproducibility of renewables generation')
@click.option('--seed-for-dispatch', default=None, help='Input seed to ensure reproducibility of dispatch')
@click.option('--ignore-warnings', is_flag=True,
              help='Ignore the warnings related to the existence of data files '
                   'in the chosen output directory.')
@click.option('--scenario_name', default='', help='subname to add to the generated scenario output folder, as Scenario_subname_i')
@click.option('--nb_core', default=1, help='number of cores to parallelize the number of scenarios')
def generate_mp(case, start_date, weeks, by_n_weeks, n_scenarios, mode,
             input_folder, output_folder, scenario_name,
             seed_for_loads, seed_for_res, seed_for_dispatch, nb_core, ignore_warnings):
    generate_mp_core(case, start_date, weeks, by_n_weeks, n_scenarios, mode,
                     input_folder, output_folder, scenario_name,
                     seed_for_loads, seed_for_res, seed_for_dispatch, nb_core, ignore_warnings)


def generate_mp_core(case, start_date, weeks, by_n_weeks, n_scenarios, mode,
             input_folder, output_folder, scenario_name,
             seed_for_loads, seed_for_res, seed_for_dispatch, nb_core, ignore_warnings):

    start_time = time.time()
    print(case)
    # create folders

    # get scenario name ids
    scenario_base_name = cst.SCENARIO_FOLDER_BASE_NAME
    if scenario_name:
        scenario_base_name = '_'.join([scenario_base_name, str(scenario_name)])

    scen_names = gu.folder_name_pattern(scenario_base_name, n_scenarios)

    generation_output_folder, kpi_output_folder = create_directory_tree(
        case, start_date, output_folder, scenario_base_name, n_scenarios, mode,
        warn_user=not ignore_warnings)

    # seeds
    default_seed = generate_default_seed()
    seed_for_loads = parse_seed_arg(seed_for_loads, '--seed-for-loads',
                                    default_seed)
    seed_for_res = parse_seed_arg(seed_for_res, '--seed-for-res',
                                  default_seed)
    seed_for_dispatch = parse_seed_arg(seed_for_dispatch, '--seed-for-dispatch',
                                       default_seed)

    initial_seeds = dict(
        loads=seed_for_loads,
        renewables=seed_for_res,
        dispatch=seed_for_dispatch
    )

    print('initial_seeds')
    print(initial_seeds)
    dump_seeds(generation_output_folder, initial_seeds, scenario_name)

    if n_scenarios >= 2:
        seeds_for_loads, seeds_for_res, seeds_for_disp = gu.generate_seeds(
            n_scenarios, seed_for_loads, seed_for_res, seed_for_dispatch
        )
    else:
        # in case uou want to reproduce a specific scenario already generated
        # with the seeds to consider
        seeds_for_loads = [seed_for_loads]
        seeds_for_res = [seed_for_res]
        seeds_for_disp = [seed_for_dispatch]

    # multi-processing
    pool = multiprocessing.Pool(nb_core)
    iterable = [i for i in range(n_scenarios)]
    multiprocessing_func = partial(
        generate_per_scenario,
        case, start_date, weeks, by_n_weeks, mode, input_folder,
        kpi_output_folder, generation_output_folder, scen_names,
        seeds_for_loads, seeds_for_res, seeds_for_disp, ignore_warnings)

    pool.map(multiprocessing_func, iterable)
    pool.close()
    print('multiprocessing done')
    print('Time taken = {} seconds'.format(time.time() - start_time))


def generate_per_scenario(case, start_date, weeks, by_n_weeks, mode,
             input_folder, kpi_output_folder, generation_output_folder, scen_names,
             seeds_for_loads, seeds_for_res, seeds_for_dispatch, ignore_warnings, scenario_id):
    
    n_scenarios_sub_p = 1  # one scenario to compute per process``
    scenario_name = scen_names(scenario_id)

    # get scenario seeds
    seed_for_loads = seeds_for_loads[scenario_id]
    seed_for_res = seeds_for_res[scenario_id]
    seed_for_dispatch = seeds_for_dispatch[scenario_id]
    
    scenario_seeds = dict(
        loads=seed_for_loads,
        renewables=seed_for_res,
        dispatch=seed_for_dispatch
    )
    print('seeds for scenario: '+scenario_name)
    print(scenario_seeds)
    
    # dump scenario seeds
    # noScenarioDirectoryHere=''
    # generation_output_folder, kpi_output_folder = create_directory_tree(
    #    case, start_date, output_folder, noScenarioDirectoryHere,
    #    n_scenarios_sub_p, mode, warn_user=not ignore_warnings)

    scenario_path = os.path.join(generation_output_folder, scenario_name)
    print('scenario_path: '+scenario_path)
    dump_seeds(scenario_path, scenario_seeds)

    # go to generate chronics
    generate_inner(
        case, start_date, weeks, by_n_weeks, n_scenarios_sub_p, mode,
        input_folder, kpi_output_folder, generation_output_folder,
        scen_names, seed_for_loads, seed_for_res, seed_for_dispatch, scenario_id)
    

def generate_inner(case, start_date, weeks, by_n_weeks, n_scenarios, mode,
                   input_folder, kpi_output_folder, generation_output_folder,
                   scen_names, seed_for_loads, seed_for_res,
                   seed_for_dispatch, scenario_id=None):

    ut.check_scenario(n_scenarios, scenario_id)
    time_parameters = gu.time_parameters(weeks, start_date)

    generation_input_folder = os.path.join(
        input_folder, cst.GENERATION_FOLDER_NAME
    )
    kpi_input_folder = os.path.join(
        input_folder, cst.KPI_FOLDER_NAME
    )

    year = time_parameters['year']

    # Chronic generation
    if 'L' in mode or 'R' in mode:
        generator = GeneratorBackend()
        params, loads_charac, prods_charac = gen.main(generator,
            case, n_scenarios, generation_input_folder,
            generation_output_folder, scen_names, time_parameters,
            mode, scenario_id, seed_for_loads, seed_for_res, seed_for_dispatch)
        scenario_name = scen_names(scenario_id)
        if by_n_weeks is not None and 'T' in mode:
            output_processor_to_chunks(
                generation_output_folder, scenario_name, by_n_weeks,
                n_scenarios, weeks)
            write_start_dates_for_chunks(
                generation_output_folder, scenario_name, weeks, by_n_weeks,
                n_scenarios, start_date, int(params['dt']))

    # KPI formatting and computing
    if 'R' in mode and 'K' in mode and 'T' not in mode:
        # Get and format solar and wind on all timescale, then compute KPI and save plots
        wind_solar_only = True
        kpis.main(kpi_input_folder, generation_output_folder, scen_names,
                  kpi_output_folder, year, case, n_scenarios, wind_solar_only,
                  params, loads_charac, prods_charac, scenario_id)

    elif 'T' in mode and 'K' in mode:
        # Get and format dispatched chronics, then compute KPI and save plots
        wind_solar_only = False
        kpis.main(kpi_input_folder, generation_output_folder, scen_names,
                  kpi_output_folder, year, case, n_scenarios, wind_solar_only,
                  params, loads_charac, prods_charac, scenario_id)


def create_directory_tree(case, start_date, output_directory, scenario_name,
                          n_scenarios, mode, warn_user=True):
    gen_path_to_create = os.path.join(
        output_directory, cst.GENERATION_FOLDER_NAME, case, start_date)
    if warn_user and os.path.isdir(gen_path_to_create):
        gu.warn_if_output_folder_not_empty(gen_path_to_create)
    os.makedirs(gen_path_to_create, exist_ok=True)

    kpi_path_to_create = None
    if 'K' in mode:
        kpi_path_to_create = os.path.join(
            output_directory, cst.KPI_FOLDER_NAME, case, start_date)
        if warn_user and os.path.isdir(kpi_path_to_create):
            gu.warn_if_output_folder_not_empty(kpi_path_to_create)
        os.makedirs(kpi_path_to_create, exist_ok=True)

    scen_name_generator = gu.folder_name_pattern(
        scenario_name, n_scenarios)
    for i in range(n_scenarios):
        s_name = scen_name_generator(i)
        scenario_path_to_create = os.path.join(gen_path_to_create, s_name)
        os.makedirs(scenario_path_to_create, exist_ok=True)

        if 'K' in mode:
            scenario_kpi_path_to_create = os.path.join(
                kpi_path_to_create, s_name, cst.KPI_IMAGES_FOLDER_NAME
            )
            os.makedirs(scenario_kpi_path_to_create, exist_ok=True)

    return gen_path_to_create, kpi_path_to_create


if __name__ == "__main__":
    # # Default arguments for dev mode
    # case = 'case118_l2rpn_wcci' #'case118_l2rpn_wcci' #'case118_l2rpn_neurips_1x' #'case118_l2rpn_neurips_1x_GAN'
    # start_date = '2012-01-01'
    # weeks = 4
    # by_n_weeks = 4
    # n_scenarios = 1
    # mode = 'LRT'
    # input_folder = 'getting_started/example/input' #'getting_started/example/input' #'input_data'
    # output_folder =  'getting_started/example/output' #'getting_started/example/output' #'output' #'output_gan'
    # scenario_name = "" #"year_wind_solar" "january_wind_solar_dispatch"
    # seed_for_loads = 912206665
    # seed_for_res = 912206665
    # seed_for_dispatch = 912206665
    # nb_core = 1
    # ignore_warnings = True
    #
    # # Run main function (only works with absolute path)
    # cwd = os.getcwd()
    # input_folder = os.path.join(cwd,input_folder)
    # output_folder = os.path.join(cwd, output_folder)
    # generate_mp_core(case, start_date, weeks, by_n_weeks, n_scenarios, mode,
    #                  input_folder, output_folder, scenario_name,
    #                  seed_for_loads, seed_for_res, seed_for_dispatch, nb_core, ignore_warnings)
    generate_mp()







import os
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# Other Python libraries
import pandas as pd
import numpy as np

# Libraries developed for this module
from . import solar_wind_utils as swutils
import chronix2grid.constants as cst
from chronix2grid.generation.renewable.gan_utils import ReplayedGAN, generate_gaussian_inputs, post_process_sample, load_model

import tensorflow as tf

def main_gan(scenario_destination_path, seed, params, prods_charac, write_results = True):


    np.random.seed(seed)

    network_folder = params["network_folder"]

    # Define datetime indices
    datetime_index = pd.date_range(
        start=params['start_date'],
        end=params['end_date'],
        freq=str(params['dt']) + 'min')

    # Init tensorflow graphs and sessions
    solar_graph = tf.Graph()
    wind_graph = tf.Graph()
    sess_solar = tf.InteractiveSession(graph=solar_graph)
    sess_wind = tf.InteractiveSession(graph=wind_graph)

    #### SOLAR
    with sess_solar.as_default():
        with solar_graph.as_default():
            dcgan_model_solar, sess_solar = load_model(sess_solar, params, network_folder, carrier='solar')
            solar_series = run_model(sess_solar, dcgan_model_solar, params, prods_charac, datetime_index, carrier='solar')

    #### WIND
    with sess_wind.as_default():
        with wind_graph.as_default():
            dcgan_model_wind, sess_wind = load_model(sess_wind, params, network_folder, carrier = 'wind')
            wind_series = run_model(sess_wind, dcgan_model_wind, params, prods_charac, datetime_index, carrier = 'wind')

    # Concatenate
    prods_series = pd.concat([wind_series, solar_series.loc[:, solar_series.columns != 'datetime']], axis=1)

    # Save files
    print('Saving files in zipped csv')
    if not os.path.exists(scenario_destination_path):
        os.makedirs(scenario_destination_path)

    prod_solar_forecasted =  swutils.create_csv(
        solar_series,
        os.path.join(scenario_destination_path, 'solar_p_forecasted.csv.bz2'),
        reordering=True,
        shift=True,
        write_results=write_results,
        index=False
    )

    prod_solar = swutils.create_csv(
        solar_series,
        os.path.join(scenario_destination_path, 'solar_p.csv.bz2'),
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results
    )

    prod_wind_forecasted = swutils.create_csv(
        wind_series,
        os.path.join(scenario_destination_path, 'wind_p_forecasted.csv.bz2'),
        reordering=True,
        shift=True,
        write_results=write_results,
        index=False
    )

    prod_wind = swutils.create_csv(
        wind_series, os.path.join(scenario_destination_path, 'wind_p.csv.bz2'),
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results
    )

    prod_p = swutils.create_csv(
        prods_series, os.path.join(scenario_destination_path, 'prod_p.csv.bz2'),
        reordering=True,
        noise=params['planned_std'],
        write_results=write_results
    )

    prod_v = prods_charac[['name', 'V']].set_index('name')
    prod_v = prod_v.T
    prod_v.index = [0]
    prod_v = prod_v.reindex(range(len(prod_p)))
    prod_v = prod_v.fillna(method='ffill') * 1.04

    prod_v.to_csv(
        os.path.join(scenario_destination_path, 'prod_v.csv.bz2'),
        sep=';',
        index=False,
        float_format=cst.FLOATING_POINT_PRECISION_FORMAT
    )

    return prod_solar, prod_solar_forecasted, prod_wind, prod_wind_forecasted

def run_model(sess, dcgan_model, params, prods_charac, datetime_index, carrier):
    # Generate random inputs
    Y, Z = generate_gaussian_inputs(params, carrier)

    # Simulate with GAN
    Z_tf_sample, Y_tf_sample, image_tf_sample = dcgan_model.samples_generator(batch_size=params["batch_size_"+carrier])
    generated_batches = []
    for y, z in zip(Y, Z):
        generated_samples = sess.run(
            image_tf_sample,
            feed_dict={
                Z_tf_sample: z,
                Y_tf_sample: y
            })
        generated_batches.append(generated_samples)
    series = post_process_sample(generated_batches, params, prods_charac, datetime_index, carrier=carrier)
    return series



import os
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from itertools import product
from pathlib import Path
from matplotlib import pyplot as plt

class NuclearKPI:

    def __init__(self, kpi_databanks):
        self.kpi_databanks = kpi_databanks

    def nuclear_kpi(self, save_plots=False):

        """
        Return:
        ------

        None
        """

        # Get the nuclear gen names
        nuclear_filter = self.prod_charac.type.isin(['nuclear'])
        nuclear_names = self.prod_charac.name.loc[nuclear_filter].values

        # Extract only nuclear power plants
        nuclear_ref = self.ref_dispatch[nuclear_names]
        nuclear_syn = self.syn_dispatch[nuclear_names]

        # Plot results
        fig, axes = plt.subplots(1, 2, figsize=(14, 5))
        axes[0].hist(nuclear_ref.values, bins=100)
        axes[1].hist(nuclear_syn.values, bins=100)
        axes[0].set_title('Nuclear Reference Distribution')
        axes[1].set_title('Nuclear Synthetic Distribution')

        if save_plots:
            # Save plot as pnd
            extent0 = axes[0].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            extent1 = axes[1].get_window_extent().transformed(fig.dpi_scale_trans.inverted())
            fig.savefig('images/nuclear_kpi/ref_dispatch_histogram.png', bbox_inches=extent0.expanded(1.3, 1.3))
            fig.savefig('images/nuclear_kpi/syn_dispatch_histogram.png', bbox_inches=extent1.expanded(1.3, 1.3))

        return None



from PypsaDispatchBackend.EDispatch_L2RPN2020 import run_economic_dispatch # TODO: Supprimer cette dépendance car pas utile (utiliser utils dans chronix2grid)
from .dispatch_loss_utils import run_grid2op_simulation_donothing, correct_scenario_loss, move_chronics_temporarily, \
    remove_temporary_chronics, remove_simulation_data



def main(dispatcher, input_folder, output_folder, grid_folder, seed, params, params_opf):
    """

    Parameters
    ----------
    dispatcher : Dispatcher
        The Dispatcher instance used for running the OPF
    input_folder : str
        The path to the directory containing the inputs for the dispatch
    output_folder
        The path of the directory that will receive the outputs of the dispatch
    grid_folder
        The path of the directory that contains grid information (used only if loss simulation are activated)
    seed : int
        Random seed for parallel execution
    params_opf : dict
        Options for the OPF

    Returns
    -------
    DispatchResults
        The namedtuple return by Dispatcher.run method
    """

    #np.random.seed(seed) # already done before

    hydro_constraints = dispatcher.make_hydro_constraints_from_res_load_scenario()
    agg_load_without_renew = dispatcher.net_load(params_opf['losses_pct'],
                                                 name=dispatcher.loads.index[0])

    dispatch_results = dispatcher.run(
        agg_load_without_renew,
        params=params_opf,
        gen_constraints=hydro_constraints,
        ramp_mode=parse_ramp_mode(params_opf['ramp_mode']),
        by_carrier=params_opf['dispatch_by_carrier'],
        pyomo=params_opf['pyomo'],
        solver_name=params_opf['solver_name']
    )
    dispatcher.save_results(params, output_folder)

    if params_opf["loss_grid2op_simulation"]:
        new_prod_p, new_prod_forecasted_p = simulate_loss(grid_folder, output_folder, params_opf, write_results = True)
        dispatch_results = update_results_loss(dispatch_results, new_prod_p, params_opf)
    return dispatch_results

def update_results_loss(dispatch_results, new_prod_p, params_opf):
    dispatch_results[0].prods_dispatch[params_opf['nameSlack']] = new_prod_p[params_opf['nameSlack']]
    return dispatch_results

def simulate_loss(input_folder, output_folder, params_opf, write_results = True):
    scenario_folder_path = output_folder
    grid_folder_g2op = input_folder
    move_chronics_temporarily(scenario_folder_path, grid_folder_g2op)
    # try:
    run_grid2op_simulation_donothing(grid_folder_g2op, scenario_folder_path,
                                 agent_type=params_opf['agent_type'])
    # except RuntimeError:
    #     remove_temporary_chronics(grid_folder_g2op)
    #     raise RuntimeError("Error in Grid2op simulation, temporary folder deleted")
    dispatch_results_corrected = correct_scenario_loss(scenario_folder_path, scenario_folder_path, params_opf, grid_folder_g2op)
    remove_temporary_chronics(grid_folder_g2op)
    #remove_simulation_data(scenario_folder_path)
    return dispatch_results_corrected


def parse_ramp_mode(mode):
    """
    Parse a string representing the difficulty of the ramps in the OPF into
    a RampMode Enum value
    Parameters
    ----------
    mode : str
        The difficulty mode for the ramps in the OPF

    Returns
    -------
    RampMode
        The encoded RampMode value

    """
    if mode == 'hard':
        return run_economic_dispatch.RampMode.hard
    if mode == 'medium':
        return run_economic_dispatch.RampMode.medium
    if mode == 'easy':
        return run_economic_dispatch.RampMode.easy
    if mode == '':
        return run_economic_dispatch.RampMode.none
    raise ValueError(f'mode only takes values from (hard, medium, easy, none), '
                     '{mode} was passed')



"""Class for the economic dispatch framework. Allows to parametrize and run
an economic dispatch based on RES and consumption time series"""

from collections import namedtuple
import os

import grid2op
from grid2op.Chronics import ChangeNothing
import pandas as pd
import pypsa

from .EDispatch_L2RPN2020.run_economic_dispatch import main_run_disptach

## Dépendances à Chronix2Grid
from chronix2grid.generation.dispatch.EconomicDispatch import Dispatcher
from chronix2grid.generation.dispatch.utils import RampMode

DispatchResults = namedtuple('DispatchResults', ['chronix', 'terminal_conditions'])


class PypsaDispatcher(Dispatcher, pypsa.Network):
    """
    Inheriting from Dispatcher to implement abstract methods thanks to Pypsa API
    Wrapper around a pypsa.Network to add higher level methods
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.add('Bus', 'node')
        self.add('Load', name='agg_load', bus='node')
        self._env = None  # The grid2op environment when instanciated with from_gri2dop_env
        self._chronix_scenario = None
        self._simplified_chronix_scenario = None
        self._has_results = False
        self._has_simplified_results = False

    @classmethod
    def from_gri2op_env(cls, grid2op_env):
        """
        Implements the abstract method of *Dispatcher*

        Parameters
        ----------
        grid2op_env

        Returns
        -------
        net: :class:`pypsa.Network`
        """
        net = cls()
        net._env = grid2op_env

        carrier_types_to_exclude = ['wind', 'solar']

        # PATCH
        # to avoid problems for respecting pmax and ramps when rounding production values in chronics at the end, we apply a correcting factor
        PmaxCorrectingFactor = 1
        RampCorrectingFactor = 0.1

        for i, generator in enumerate(grid2op_env.name_gen):
            gen_type = grid2op_env.gen_type[i]
            if gen_type not in carrier_types_to_exclude:
                p_max = grid2op_env.gen_pmax[i]
                pnom = p_max - PmaxCorrectingFactor
                rampUp = (grid2op_env.gen_max_ramp_up[i] - RampCorrectingFactor) / p_max
                RampDown = (grid2op_env.gen_max_ramp_down[i] - RampCorrectingFactor) / p_max

                net.add(
                    class_name='Generator', name=generator, bus='node',
                    p_nom=pnom, carrier=grid2op_env.gen_type[i],
                    marginal_cost=grid2op_env.gen_cost_per_MW[i],
                    ramp_limit_up=rampUp,
                    ramp_limit_down=RampDown,
                )

        return net


    def run(self, load, params, gen_constraints=None,
            ramp_mode=RampMode.hard, by_carrier=False, **kwargs):
        """
        Implements the abstract method of *Dispatcher*

        Returns
        -------
        simplified_net: :class:`pypsa.Network`
        """
        prods_dispatch, terminal_conditions, marginal_prices = main_run_disptach(
            self if not by_carrier else self.simplify_net(),
            load, params, gen_constraints, ramp_mode, **kwargs)
        if by_carrier:
            self._simplified_chronix_scenario = self._chronix_scenario.simplify_chronix()
            self._simplified_chronix_scenario.prods_dispatch = prods_dispatch
            self._simplified_chronix_scenario.marginal_prices = marginal_prices
            results = self._simplified_chronix_scenario
            self._has_simplified_results = True
            self._has_results = False
        else:
            self._chronix_scenario.prods_dispatch = prods_dispatch
            self._chronix_scenario.marginal_prices = marginal_prices
            results = self._chronix_scenario
            self._has_results = True
            self._has_simplified_results = False
        self.reset_ramps_from_grid2op_env()
        return DispatchResults(chronix=results, terminal_conditions=terminal_conditions)

    def simplify_net(self):
        """
        Implements the abstract method of *Dispatcher*
        """
        carriers = self.generators.carrier.unique()
        simplified_net = PypsaDispatcher()
        for carrier in carriers:
            names = self.generators[self.generators.carrier == carrier].index.tolist()

            gens = self.generators.loc[
                names,
                ['p_nom', 'ramp_limit_up', 'ramp_limit_down', 'marginal_cost']
            ]
            gens['ramp_up_mw'] = gens['p_nom'] * gens['ramp_limit_up']
            gens['ramp_down_mw'] = gens['p_nom'] * gens['ramp_limit_down']
            params = gens.agg(['sum', 'mean'])
            simplified_net.add(
                class_name='Generator', name=carrier, bus='node',
                p_nom=params.loc['sum', 'p_nom'], carrier=carrier,
                marginal_cost=params.loc['mean', 'marginal_cost'],
                ramp_limit_up=params.loc['sum', 'ramp_up_mw'] / params.loc['sum', 'p_nom'],
                ramp_limit_down=params.loc['sum', 'ramp_down_mw'] / params.loc['sum', 'p_nom'],
            )
        simplified_net._hydro_file_path = self._hydro_file_path
        simplified_net._min_hydro_pu = self._min_hydro_pu.iloc[:, 0]
        simplified_net._max_hydro_pu = self._max_hydro_pu.iloc[:, 0]

        print('simplified dispatch by carrier')
        full_ramp = simplified_net.generators['p_nom'] * simplified_net.generators['ramp_limit_up']
        df_full_ramp = pd.DataFrame({'full_ramp': full_ramp})

        print(pd.concat(
            [
                simplified_net.generators[['p_nom', 'ramp_limit_up',
                                           'ramp_limit_down', 'marginal_cost']],
                df_full_ramp
            ], axis=1))
        return simplified_net



if __name__ == "__main__":

    INPUT_FOLDER = 'chronix2grid/generation/input'
    CASE = 'case118_l2rpn'
    path_grid = os.path.join(INPUT_FOLDER, CASE)

    losses_pct = 3.0

    env118_blank = grid2op.make(
        test=True,
        grid_path=os.path.join(path_grid, "L2RPN_2020_case118_redesigned.json"),
        chronics_class=ChangeNothing,
    )
    params = {'snapshots': [],
              'step_opf_min': 5,
              'mode_opf': 'week',
              'reactive_comp': 1.025,
              }
    chronics_path_gen = os.path.join(INPUT_FOLDER, "dispatch", str(2012))
    this_path = os.path.join(chronics_path_gen, 'Scenario_0')
    dispatch = PypsaDispatcher.from_gri2op_env(env118_blank)
    dispatch.read_hydro_guide_curves(os.path.join(INPUT_FOLDER, 'patterns', 'hydro.csv'))
    dispatch.read_load_and_res_scenario(os.path.join(this_path, 'load_p.csv.bz2'),
                                        os.path.join(this_path, 'prod_p.csv.bz2'),
                                        'Scenario_0')
    dispatch.make_hydro_constraints_from_res_load_scenario()
    net_by_carrier = dispatch.simplify_net()
    agg_load_without_renew = dispatch.net_load(losses_pct, name=dispatch.loads.index[0])

    # Prepare gen constraints for EDispatch module
    hydro_constraints = {'p_max_pu': dispatch._max_hydro_pu.copy(),
                         'p_min_pu': dispatch._min_hydro_pu.copy()}

    opf_dispatch, term_conditions = dispatch.run(
        agg_load_without_renew,
        params=params,
        gen_constraints=hydro_constraints,
        ramp_mode=run_economic_dispatch.RampMode.easy,
        by_carrier=False  # True to run the dispatch only aggregated generators by carrier
    )

    dispatch.save_results(params, '.')
    test_prods = pd.read_csv('./Scenario_0/prod_p.csv.bz2', sep=";")
    test_prices = pd.read_csv('./Scenario_0/prices.csv.bz2', sep=";")



"""Class for the economic dispatch framework. Allows to parametrize and run
an economic dispatch based on RES and consumption time series"""

import os
from abc import ABC, abstractmethod
from copy import deepcopy
import datetime as dt
from collections import namedtuple

import grid2op
from grid2op.Chronics import ChangeNothing
import pandas as pd
import plotly.express as px

from chronix2grid.generation.dispatch.utils import RampMode, add_noise_gen, modify_hydro_ramps, modify_slack_characs
import chronix2grid.constants as cst

DispatchResults = namedtuple('DispatchResults', ['chronix', 'terminal_conditions'])

def init_dispatcher_from_config(grid_path, input_folder, dispatcher_class, params_opf):
    # Read grid and gens characs
    env118_withoutchron = grid2op.make("rte_case118_example",
                                       test=True,
                                       grid_path=grid_path,
                                       chronics_class=ChangeNothing)
    # env118_withoutchron = grid2op.make(grid_path)

    # Generators temporary adjusts
    env118_withoutchron = modify_hydro_ramps(env118_withoutchron, params_opf["hydro_ramp_reduction_factor"])

    if params_opf["slack_p_max_reduction"] != 0. or params_opf['slack_ramp_max_reduction'] != 0.:
        env118_withoutchron = modify_slack_characs(env118_withoutchron,
                                                 params_opf["nameSlack"],
                                                 params_opf["slack_p_max_reduction"],
                                                 params_opf["slack_ramp_max_reduction"])

    # Dispatcher object init
    dispatcher = dispatcher_class.from_gri2op_env(env118_withoutchron)

    dispatcher.read_hydro_guide_curves(
        os.path.join(input_folder, 'patterns', 'hydro_french.csv'))
    return dispatcher



class Dispatcher(ABC):
    """Abstract Class Dispatcher with some operational method that don't depend on the dispatcher technology (pypsa...)
    and some others that are implemented specifically in inheriting classes"""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def run(self, load, params, gen_constraints=None,
            ramp_mode=RampMode.hard, by_carrier=False, **kwargs):
        """Run the proper dispatch optimization. Have to be implemented in inheriting classes"""

    @abstractmethod
    def simplify_net(self):
        pass

    @classmethod
    @abstractmethod
    def from_gri2op_env(cls, grid2op_env):
        """Reads grid features from a grid2op environment into a specific object.
        Have to be implemented in inheriting classes according to the type of model"""

    @property
    def wind_p(self):
        if self._chronix_scenario is None:
            raise Exception('Cannot access this property before instantiated the Load and '
                            'renewables scenario.')
        return self._chronix_scenario.wind_p

    @property
    def solar_p(self):
        if self._chronix_scenario is None:
            raise Exception('Cannot access this property before instantiated the Load and '
                            'renewables scenario.')
        return self._chronix_scenario.solar_p

    @property
    def chronix_scenario(self):
        """

        Returns
        -------
        chronix_scenario: :class:`chronix2grid.generation.dispatch.EconomicDispatch.ChroniXScenario`
            Object that contains all the chronics that are relevant to perform a dispatch
        """
        if self._chronix_scenario is None:
            raise Exception('Cannot access this property before instantiated the Load and '
                            'renewables scenario.')
        return self._chronix_scenario

    @chronix_scenario.setter
    def chronix_scenario(self, chronix_scenario):
        if not isinstance(chronix_scenario, ChroniXScenario):
            raise Exception('The chronix_scenario argument should be an instance of '
                            'ChronixScenario.')
        self._chronix_scenario = chronix_scenario

    def net_load(self, losses_pct, name):
        if self._chronix_scenario is None:
            raise Exception('Cannot compute net load before instantiated the Load and'
                            'renewables scenario.')
        return self._chronix_scenario.net_load(losses_pct, name)

    def nlargest_ramps(self, n, losses_pct):
        ramps = self.net_load(losses_pct, "").diff()
        return ramps.nlargest(n, ramps.columns[0])

    def reset_ramps_from_grid2op_env(self):
        if self._env is None:
            raise Exception('This method can only be applied when Dispatch has been'
                            'instantiated from a grid2op Environment.')
        for i, generator in enumerate(self._env.name_gen):
            if generator in self.generators.index:
                self.generators.loc[generator, 'ramp_limit_up'] = \
                    self._env.gen_max_ramp_up[i] / self._env.gen_pmax[i]
                self.generators.loc[generator, 'ramp_limit_down'] = \
                    self._env.gen_max_ramp_down[i] / self._env.gen_pmax[i]

    def read_hydro_guide_curves(self, hydro_file_path):
        """
        Reads realistic hydro pattern that provides seasonal boundaries to the hydro production.
        This constraint in the dispatch problem leads to more realistic hydro production

        Parameters
        ----------
        hydro_file_path: ``str``

        """
        dateparse = lambda x: dt.datetime.strptime(x, '%Y-%m-%d %H:%M')
        hydro_pattern = pd.read_csv(hydro_file_path, usecols=[0, 2, 3],
                                    parse_dates=[0], date_parser=dateparse)
        hydro_pattern.set_index(hydro_pattern.columns[0], inplace=True)
        hydro_names = self.generators[self.generators.carrier == 'hydro'].index

        for extremum in ['min', 'max']:
            hydro_pu = hydro_pattern[[f'p_{extremum}_u'] * len(hydro_names)]
            hydro_pu.columns = hydro_names
            hydro_pu.index = hydro_pu.index.map(
                lambda x: (x.month, x.day, x.hour, x.minute, x.second))
            setattr(self, f'_{extremum}_hydro_pu', hydro_pu)

        self._hydro_file_path = hydro_file_path

    def read_load_and_res_scenario(self, load_path_file, prod_path_file,
                                   scenario_name, start_date, end_date, dt):
        if self._env is None:
            raise Exception('This method can only be applied when Dispatch has been'
                            'instantiated from a grid2op Environment.')
        res_names = dict(
            wind=[name for i, name in enumerate(self._env.name_gen)
                  if self._env.gen_type[i] == 'wind'],
            solar=[name for i, name in enumerate(self._env.name_gen)
                   if self._env.gen_type[i] == 'solar']
        )
        self._chronix_scenario = ChroniXScenario.from_disk(
            load_path_file, prod_path_file,
            res_names, scenario_name=scenario_name,
            start_date=start_date, end_date=end_date, dt=dt
        )

    def make_hydro_constraints_from_res_load_scenario(self):
        if self._chronix_scenario is None or self._hydro_file_path is None:
            raise Exception('This method can only be applied when a Scenario for load'
                            'and renewables has been instantiated and hydro guide'
                            'curves have been read.')
        index_slice = self._chronix_scenario.loads.index.map(
            lambda x: (x.month, x.day, x.hour, x.minute, x.second)
        )

        self._min_hydro_pu = self._min_hydro_pu.reindex(index_slice).fillna(method='ffill')
        self._max_hydro_pu = self._max_hydro_pu.reindex(index_slice).fillna(method='ffill')

        return {'p_max_pu': self._max_hydro_pu.copy(), 'p_min_pu': self._min_hydro_pu.copy()}

    def modify_marginal_costs(self, new_costs):
        for carrier, new_cost in new_costs.items():
            try:
                targeted_generators = self.generators.carrier.isin([carrier])
                self.generators.loc[targeted_generators, 'marginal_cost'] = new_cost
            except KeyError:
                print(f'Carrier {carrier} is not a valid carrier')
                raise

    def plot_ramps(self):
        caract_gen = self.generators[['p_nom', 'carrier', 'ramp_limit_up']].reset_index()
        caract_gen = caract_gen.rename(columns={'index': 'name'})

        fig = px.scatter(
            caract_gen, x='p_nom', y='ramp_limit_up', color='carrier',
            hover_data=['name']
        )
        return fig

    def save_results(self, params, output_folder):
        """
        Saves dispatch results in prod_p.csv.bz2, prod_p_forecasted.csv.bz2, load_p.csv.bz2, prices.csv

        Parameters
        ----------
        params: ``dict``
        output_folder: ``str``

        """
        if not self._has_results and not self._has_simplified_results:
            print('The optimization has first to run successfully in order to '
                  'save results.')
        if self._has_results:
            print('Saving results for the grids with individual generators...')
            res_load_scenario = self.chronix_scenario
        else:
            print('Saving results for the grids with aggregated generators by carriers...')
            res_load_scenario = self._simplified_chronix_scenario

        full_opf_dispatch = pd.concat(
            [res_load_scenario.prods_dispatch, res_load_scenario.wind_p,
             res_load_scenario.solar_p],
            axis=1
        )
        try:
            full_opf_dispatch = full_opf_dispatch[self._env.name_gen].round(2)
        except KeyError:
            # Either we're trying to save results from a simplified dispatch or
            # using the save function before instanciating an env.
            pass

        gen_cap = pd.Series({gen_name: gen_pmax for gen_name, gen_pmax in
                             zip(self._env.name_gen, self._env.gen_pmax)})

        prod_p_forecasted_with_noise = add_noise_gen(full_opf_dispatch, gen_cap, noise_factor=params['planned_std'])

        # prod_p_forecasted_with_noise.to_csv(
        prod_p_forecasted_with_noise.to_csv(
            os.path.join(output_folder, "prod_p_forecasted.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        # prod_p_with_noise.to_csv(
        full_opf_dispatch.to_csv(
            os.path.join(output_folder, "prod_p.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        res_load_scenario.marginal_prices.to_csv(
            os.path.join(output_folder, "prices.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )
        res_load_scenario.loads.to_csv(
            os.path.join(output_folder, "load_p.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )

class ChroniXScenario:
    def __init__(self, loads, prods, res_names, scenario_name, loss=None):
        self.loads = loads
        self.wind_p = prods[res_names['wind']]
        self.solar_p = prods[res_names['solar']]
        self.total_res = pd.concat([self.wind_p, self.solar_p], axis=1).sum(axis=1)
        self.prods_dispatch = None  # Will receive the results of the dispatch
        self.marginal_prices = None  # Will receive the marginal prices associated to a dispatch
        self.name = scenario_name
        self.loss = loss

    @classmethod
    def from_disk(cls, load_path_file, prod_path_file, res_names, scenario_name,
                  start_date, end_date, dt, loss_path_file=None):
        loads = pd.read_csv(load_path_file, sep=';')
        prods = pd.read_csv(prod_path_file, sep=';')
        if loss_path_file is not None:
            loss = pd.read_csv(loss_path_file, sep=';')
        else:
            loss = None
        datetime_index = pd.date_range(
            start=start_date,
            end=end_date,
            freq=str(dt) + 'min')
        loads.index = datetime_index[:len(loads)]
        prods.index = datetime_index[:len(prods)]
        return cls(loads, prods, res_names, scenario_name, loss)

    def net_load(self, losses_pct, name):
        if self.loss is None:
            load_minus_losses = self.loads.sum(axis=1) * (1 + losses_pct / 100)
        else:
            load_minus_losses = self.loads.sum(axis=1) + self.loss
        return (load_minus_losses - self.total_res).to_frame(name=name)

    def simplify_chronix(self):
        simplified_chronix = deepcopy(self)
        simplified_chronix.wind_p = simplified_chronix.wind_p.sum(axis=1).to_frame(name='wind')
        simplified_chronix.solar_p = simplified_chronix.solar_p.sum(axis=1).to_frame(name='solar')
        simplified_chronix.name = simplified_chronix.name + 'by_carrier'
        return simplified_chronix


# if __name__ == "__main__":
#
#     INPUT_FOLDER = 'chronix2grid/generation/input'
#     CASE = 'case118_l2rpn'
#     path_grid = os.path.join(INPUT_FOLDER, CASE)
#
#     losses_pct = 3.0
#
#     env118_blank = grid2op.make(
#         test=True,
#         grid_path=os.path.join(path_grid, "L2RPN_2020_case118_redesigned.json"),
#         chronics_class=ChangeNothing,
#     )
#     params = {'snapshots': [],
#               'step_opf_min': 5,
#               'mode_opf': 'week',
#               'reactive_comp': 1.025,
#               }
#     chronics_path_gen = os.path.join(INPUT_FOLDER, "dispatch", str(2012))
#     this_path = os.path.join(chronics_path_gen, 'Scenario_0')
#     dispatch = Dispatcher.from_gri2op_env(env118_blank)
#     dispatch.read_hydro_guide_curves(os.path.join(INPUT_FOLDER, 'patterns', 'hydro.csv'))
#     dispatch.read_load_and_res_scenario(os.path.join(this_path, 'load_p.csv.bz2'),
#                                         os.path.join(this_path, 'prod_p.csv.bz2'),
#                                         'Scenario_0')
#     dispatch.make_hydro_constraints_from_res_load_scenario()
#     net_by_carrier = dispatch.simplify_net()
#     agg_load_without_renew = dispatch.net_load(losses_pct, name=dispatch.loads.index[0])
#
#     # Prepare gen constraints for EDispatch module
#     hydro_constraints = {'p_max_pu': dispatch._max_hydro_pu.copy(),
#                          'p_min_pu': dispatch._min_hydro_pu.copy()}
#
#     opf_dispatch, term_conditions = dispatch.run(
#         agg_load_without_renew,
#         params=params,
#         gen_constraints=hydro_constraints,
#         ramp_mode=run_economic_dispatch.RampMode.easy,
#         by_carrier=False  # True to run the dispatch only aggregated generators by carrier
#     )
#
#     dispatch.save_results(params, '.')
#     test_prods = pd.read_csv('./Scenario_0/prod_p.csv.bz2', sep=";")
#     test_prices = pd.read_csv('./Scenario_0/prices.csv.bz2', sep=";")



import os
import csv
import sys
import numpy as np
import pandas as pd

import chronix2grid.constants as cst

def usa_gan_trainingset_to_kpi(kpi_case_input_folder, timestep, prods_charac, loads_charac, params,year):
    try:
        timestep_int = int(timestep[:3])
        if timestep_int != 120:
            raise ValueError('GAN requires timestep of 120 min to compute KPI. Please change timestep in paramsKPI.json')
    except:
        raise ValueError('GAN requires timestep of 120 min to compute KPI. Please change timestep in paramsKPI.json')


    # Selecting desired time index
    start_date = params["start_date"]
    end_date = params['end_date']
    datetime_index = pd.date_range(
        start=start_date,
        end=end_date,
        freq=str(timestep))
    T = len(datetime_index)


    ## WIND
    repo_in = os.path.join(kpi_case_input_folder, 'USA - Washington State', 'NREL',
                           cst.GAN_TRAINING_SET_REFERENCE_FOLDER, "wind.csv")
    ref_prod_wind = gan_trainingdata_processing_wind(repo_in, T, prods_charac, params)

    ## Solar TEMPORARY ============
    repo_in = os.path.join(kpi_case_input_folder, 'USA - Washington State', 'NREL',
                           cst.GAN_TRAINING_SET_REFERENCE_FOLDER, "solar.csv")
    ref_prod_solar = gan_trainingdata_processing_solar(repo_in, T, prods_charac, params)

    ## Concatenate
    ref_prod_wind.index = ref_prod_solar.index
    ref_prod = pd.concat([ref_prod_wind, ref_prod_solar], axis = 1)

    ## Load TEMPORARY ============
    load_nodes = loads_charac['name'].unique()
    ref_load = pd.DataFrame(np.ones((T,len(load_nodes)))*25, columns = load_nodes)

    ## Time index
    ref_prod['Time'] = datetime_index
    ref_prod.index = ref_prod['Time']
    ref_load['Time'] = datetime_index
    ref_load.index = ref_load['Time']
    return ref_prod, ref_load

def gan_trainingdata_processing_solar(repo_in, T, prods_charac, params):
    carrier = 'solar'
    # Loading and reshaping raw data from NREL (State of Washington wind and solar farms)
    with open(repo_in, 'r') as csvfile:
        reader = csv.reader(csvfile)
        rows = [row for row in reader]
    rows = np.array(rows, dtype=float)
    rows=rows[:104832,:] #Change to the time points in your own dataset
    # print(np.shape(rows))
    trX = np.reshape(rows.T,(-1, params['n_timesteps_' + carrier], params['n_gens_' + carrier])) #Corresponds to the GAN input dimensions.
    #print(np.shape(trX))
    m = np.ndarray.max(rows)
    #print("maximum value of solar power", m)
    trX=trX/m

    # Getting wind farms generations
    wind_gens = prods_charac[prods_charac['type'] == carrier]['name'].unique()
    N = len(wind_gens)
    N_learning = params['n_gens_solar']
    if N % 2 == 0:
        selection_gens = [i for i in range(N // 2)] + [i for i in range((N_learning-N//2), N_learning)]
    else:
        selection_gens = [i for i in range((N // 2) + 1)] + [i for i in range((N_learning-N//2), N_learning)]

    ref_prod_wind = pd.DataFrame(columns=wind_gens)
    for day in range(T // params["n_timesteps_" + carrier]):
        day_production = np.transpose(trX[day, :, selection_gens]) # np.transpose(expr) si on inverse les gens et les timesteps
        new_df = pd.DataFrame(day_production, columns=wind_gens)
        ref_prod_wind = pd.concat([ref_prod_wind, new_df], axis=0)
    day_production = np.transpose(trX[(day + 1), :(T % params["n_timesteps_" + carrier]), selection_gens]) # np.transpose(expr) si on inverse les gens et les timesteps
    new_df = pd.DataFrame(day_production, columns=wind_gens)
    ref_prod_wind = pd.concat([ref_prod_wind, new_df], axis=0)

    # Scale with Pmax
    for gen in wind_gens:
        Pmax = prods_charac.loc[prods_charac['name'] == gen, 'Pmax'].values[0]
        ref_prod_wind[gen] = ref_prod_wind[gen] * Pmax
    return ref_prod_wind

def gan_trainingdata_processing_wind(repo_in, T, prods_charac, params):
    carrier = 'wind'
    # Loading and reshaping raw data from NREL (State of Washington wind farms)
    with open(repo_in, 'r') as csvfile:
        reader = csv.reader(csvfile)
        rows = [row for row in reader]
    rows = np.array(rows, dtype=float)
    trX = []
    # print(np.shape(rows))
    m = np.ndarray.max(rows)
    # print("Maximum value of wind", m)
    # print(np.shape(rows))
    for x in range(rows.shape[1]):
        train = rows[:-288, x].reshape(-1, params['n_timesteps_' + carrier], params['n_gens_' + carrier])  # DIMENSIONS
        train = train / 16  # Pmax in data

        # print(shape(train))
        if trX == []:
            trX = train
        else:
            trX = np.concatenate((trX, train), axis=0)
    # print("Shape TrX", np.shape(trX))

    # Getting wind farms generations
    wind_gens = prods_charac[prods_charac['type'] == carrier]['name'].unique()
    N = len(wind_gens)
    N_learning = params['n_gens_wind']
    if N % 2 == 0:
        selection_gens = [i for i in range(N // 2)] + [i for i in range((N_learning-N//2), N_learning)]
    else:
        selection_gens = [i for i in range((N // 2) + 1)] + [i for i in range((N_learning-N//2), N_learning)]

    ref_prod_wind = pd.DataFrame(columns=wind_gens)
    for day in range(T // params["n_timesteps_" + carrier]):
        day_production = np.transpose(trX[day, :, selection_gens]) # np.transpose(expr) si on inverse les gens et les timesteps
        new_df = pd.DataFrame(day_production, columns=wind_gens)
        ref_prod_wind = pd.concat([ref_prod_wind, new_df], axis=0)
    day_production = np.transpose(trX[(day + 1), :(T % params["n_timesteps_" + carrier]), selection_gens]) # np.transpose(expr) si on inverse les gens et les timesteps
    new_df = pd.DataFrame(day_production, columns=wind_gens)
    ref_prod_wind = pd.concat([ref_prod_wind, new_df], axis=0)

    # Scale with Pmax
    for gen in wind_gens:
        Pmax = prods_charac.loc[prods_charac['name'] == gen, 'Pmax'].values[0]
        ref_prod_wind[gen] = ref_prod_wind[gen] * Pmax
    return ref_prod_wind

def eco2mix_to_kpi_regional(kpi_input_folder, timestep, prods_charac, loads_charac, year, params, corresp_regions):
    # Initialize dataframes to fill
    print("Importing and formatting data downloaded from regional eco2mix data")
    conso_ = pd.DataFrame()
    prod = pd.DataFrame()

    # Read data per region in corresp_regions
    for region_fictive in corresp_regions.keys():

        region = corresp_regions[region_fictive]
        repo_in = os.path.join(kpi_input_folder, 'France/eco2mix', 'eCO2mix_RTE_'+region+'_Annuel-Definitif_'+str(year)+'.csv')

        prods_charac_ = prods_charac[prods_charac['zone']==region_fictive]
        loads_charac_ = loads_charac[loads_charac['zone'] == region_fictive]

        gens = {}
        gens['solar'] = prods_charac_[prods_charac_['type'] == 'solar']['name'].values
        gens['wind'] = prods_charac_[prods_charac_['type'] == 'wind']['name'].values
        gens['hydro'] = prods_charac_[prods_charac_['type'] == 'hydro']['name'].values
        gens['nuclear'] = prods_charac_[prods_charac_['type'] == 'nuclear']['name'].values
        gens['thermal'] = prods_charac_[prods_charac_['type'] == 'thermal']['name'].values

        corresp_carrier = {'thermal':['Gaz','Fioul','Bioénergies','Charbon'],
                           'solar':['Solaire'],
                           'wind':['Eolien'],
                           'nuclear': ['Nucléaire'],
                           'hydro':['Hydraulique']}

        # Read and correct some strange values in eco2mix
        eco2mix = pd.read_csv(repo_in, sep = ';', encoding = 'latin1', decimal = ',')
        eco2mix = eco2mix.replace('ND', 0)
        eco2mix['Solaire'] = eco2mix['Solaire'].mask(eco2mix['Solaire'] < 0, 0)
        eco2mix = eco2mix.replace('-', 0)

        # Format columns names and types
        colToTake = ['Date', 'Heures', 'Consommation']
        for col in ['Fioul', 'Charbon','Gaz','Bioénergies', 'Nucléaire','Eolien', 'Solaire', 'Hydraulique']:
            if col in eco2mix.columns:
                colToTake.append(col)
                eco2mix[col] = eco2mix[col].astype(float)
        df = eco2mix[colToTake]

        # Time formatting
        df['Space'] = ' '
        df['Time'] = df['Date']+df['Space']+df['Heures']
        df['Time'] = pd.to_datetime(df['Time'], infer_datetime_format=True)
        df.set_index('Time', drop=False, inplace=True)

        # Production formatting
        for carrier_out in corresp_carrier.keys():
            agregate_carrier = np.zeros(len(df))
            for carr in corresp_carrier[carrier_out]:
                if carr in df.columns:
                    agregate_carrier += df[carr]
            df[carrier_out] = agregate_carrier

        # Equitable repartition among region generators
        for carrier in gens.keys():
            n = len(gens[carrier])
            for col in gens[carrier]:
                df[col] = df[carrier]/n

        # Resampling
        original_index = df.index
        df = df.resample(timestep).first()

        # Dropping useless columns
        agg_conso = df['Consommation']
        df.drop(columns=['Space', 'Date', 'Heures',
                         'Fioul', 'Charbon', 'Gaz', 'Bioénergies', 'Nucléaire', 'Eolien', 'Solaire', 'Hydraulique',
                         'Pompage', 'Consommation'] + list(corresp_carrier.keys()), inplace=True, errors='ignore')

        # Load computation
        loads = loads_charac_['name'].unique()

        # Equitable repartition on loads nodes
        conso = pd.DataFrame({'Time': df['Time']})
        for col in loads:
            conso[col] = agg_conso/len(loads)

        # Equalize timeline with synthetic chronics
        start_date = params['start_date']
        end_date = params['end_date']
        df = df[(df.index >= start_date) & (df.index < end_date)]
        conso = conso[(conso.index>=start_date)&(conso.index<end_date)]

        # Add in dataframe
        prod = pd.concat([prod,df], axis = 1)
        conso_ = pd.concat([conso_, conso], axis=1)

    # Read price data if exists
    repo_in = os.path.join(kpi_input_folder, 'France/eco2mix/price_'+str(year)+'.csv')
    try:
        price = pd.read_csv(repo_in, sep=';', decimal='.')
        price.index = original_index[original_index.notnull()]
        price = price.resample(timestep).first()
        price = price[(price.index >= start_date) & (price.index < end_date)]
    except:
        price = None
    return prod, conso_, price


def renewableninja_to_kpi(kpi_input_folder, timestep, loads_charac,
                          prods_charac, year, params, corresp_regions, case):

    # Initialize data frame for production
    print("Importing and formatting data downloaded from Renewable Ninja API")
    solar_file_name = f'solar.csv'
    wind_file_name = f'wind.csv'
    repo_in_solar = os.path.join(
        kpi_input_folder, cst.REFERENCE_ZONE,
        cst.RENEWABLE_NINJA_REFERENCE_FOLDER, solar_file_name)
    ninja_solar = pd.read_csv(repo_in_solar, sep=';', encoding='latin1', decimal='.')
    repo_in_wind = os.path.join(
        kpi_input_folder, cst.REFERENCE_ZONE,
        cst.RENEWABLE_NINJA_REFERENCE_FOLDER, wind_file_name)
    ninja_wind = pd.read_csv(repo_in_wind, sep=';', encoding='latin1', decimal='.')
    timestep_ninja = 60 # Pas de temps une heure dans l'extraction renewable ninja
    ninja = pd.concat([ninja_solar, ninja_wind], axis = 1)

    # Time column to add
    start_date = pd.to_datetime(str(year)+'-01-01')
    end_date = pd.to_datetime(str(year+1) + '-01-01')
    datetime_index = pd.date_range(
        start=start_date,
        end=end_date,
        freq=str(timestep_ninja) + 'min')
    datetime_index = datetime_index[:len(datetime_index)-1]

    # Resample
    ninja.index = datetime_index
    if int(timestep[:2]) < 60:
        print('Erreur: Renewable ninja requires timestep >60 min to compute KPI. Please change timestep in paramsKPI.json')
        sys.exit()
    ninja = ninja.resample(timestep).first()
    ninja['Time'] = ninja.index

    # Columns of zeros for non wind and solar generators
    gens= prods_charac[~(prods_charac['type'].isin(['solar','wind']))]['name'].values
    for gen in gens:
        ninja[gen] = 0.

    ## Electricty consumption from regional eco2mix
    conso_ = pd.DataFrame()

    for region_fictive in corresp_regions.keys():
        region = corresp_regions[region_fictive]
        repo_in = os.path.join(kpi_input_folder, 'France/eco2mix',
                               'eCO2mix_RTE_' + region + '_Annuel-Definitif_' + str(year) + '.csv')

        loads_charac_ = loads_charac[loads_charac['zone'] == region_fictive]

        eco2mix = pd.read_csv(repo_in, sep=';', encoding='latin1', decimal=',')
        colToTake = ['Date', 'Heures', 'Consommation']
        df = eco2mix[colToTake]

        # Time formatting
        df['Space'] = ' '
        df['Time'] = df['Date'] + df['Space'] + df['Heures']
        df['Time'] = pd.to_datetime(df['Time'], infer_datetime_format=True)
        df.set_index('Time', drop=False, inplace=True)

        # Correct some strange values in eco2mix data
        df = df.replace('-', 0)
        df = df.replace('ND', 0)


        # Resampling
        df = df.resample(timestep).first()
        agg_conso = df['Consommation'].astype(float)

        # Equitable repartition on loads nodes
        loads = loads_charac_['name'].unique()
        conso = pd.DataFrame({'Time': df['Time']})
        for col in loads:
            conso[col] = agg_conso / len(loads)
        conso_ = pd.concat([conso_, conso], axis = 1)

    # Equalize timeline with synthetic
    start_date = params['start_date']
    end_date = params['end_date']
    ninja = ninja[(ninja.index >= start_date) & (ninja.index < end_date)]
    conso_ = conso_[(conso_.index >= start_date) & (conso_.index < end_date)]

    return ninja, conso_


def chronics_to_kpi(chronics_repo, timestep, params, thermal = True):

    print("Importing and formatting synthetic chronics")

    # Rebuild timeline
    datetime_index = pd.date_range(
        start=params['start_date'],
        end=params['end_date'],
        freq=str(params['dt']) + 'min')

    if thermal:
        ## Format when all dispatch is generated

        # Read generated chronics after dispatch phase
        prod_p = pd.read_csv(os.path.join(chronics_repo, 'prod_p.csv.bz2'),
                             sep=';', decimal='.')
        load_p = pd.read_csv(os.path.join(chronics_repo, 'load_p.csv.bz2'),
                             sep=';', decimal='.')
        price = pd.read_csv(os.path.join(chronics_repo, 'prices.csv.bz2'),
                            sep=';', decimal='.')

        price['Time'] = datetime_index[:len(price)]

    else:
        ## Format synthetic chronics when no dispatch has been done
        solar_p = pd.read_csv(os.path.join(chronics_repo, 'solar_p.csv.bz2'), sep=';', decimal='.')
        wind_p = pd.read_csv(os.path.join(chronics_repo, 'wind_p.csv.bz2'), sep=';', decimal='.')
        prod_p = pd.concat([solar_p, wind_p], axis=1)

        load_p = pd.read_csv(os.path.join(chronics_repo, 'load_p.csv.bz2'), sep=';', decimal='.')

    prod_p['Time'] = datetime_index[:len(prod_p)]
    load_p['Time'] = datetime_index[:len(load_p)]

    # Optional resampling
    load_p['Time'] = pd.to_datetime(load_p['Time'])
    load_p.set_index('Time', drop=False, inplace=True)
    load_p = load_p.resample(timestep).first()

    prod_p['Time'] = pd.to_datetime(prod_p['Time'])
    prod_p.set_index('Time', drop=False, inplace=True)
    prod_p = prod_p.resample(timestep).first()

    # Return with price if dispatch has been made
    if not thermal:
        return prod_p, load_p
    if thermal:
        price['Time'] = pd.to_datetime(price['Time'])
        price.set_index('Time', drop=True, inplace=True)
        price = price.resample(timestep).first()
        return prod_p, load_p, price


def nrel_to_kpi(kpi_input_folder, timestep, prods_charac, loads_charac, params, year):

    # Read relation between reference generators and regions
    corresp_regions_gens = pd.read_csv(os.path.join(kpi_input_folder, 'Texas', 'corresp_gens.csv'), sep = ';')
    corresp_regions_loads = pd.read_csv(os.path.join(kpi_input_folder, 'Texas', 'corresp_loads.csv'), sep = ';')

    # Initialize data frame for production
    print("Importing and formatting data downloaded from NREL Texas")
    repo_in_prod = os.path.join(kpi_input_folder, 'Texas', 'prod_2007.csv.bz2')
    nrel_prod = pd.read_csv(repo_in_prod, sep=',', encoding='latin1', decimal='.')

    repo_in_load = os.path.join(kpi_input_folder, 'Texas', 'load_2007.csv.bz2')
    nrel_load = pd.read_csv(repo_in_load, sep=',', encoding='latin1', decimal='.')

    # Datetime processing
    nrel_prod['datetime'] = pd.to_datetime(nrel_prod['datetime'])
    nrel_load['datetime'] = pd.to_datetime(nrel_load['datetime'])

    # Test if chosen timestep is relevant for NREL
    timestep_int = int(timestep[:2])
    if timestep_int < 5:
        print('Erreur: NREL requires timestep >15 min to compute KPI. Please change timestep in paramsKPI.json')
        sys.exit()

    # For each fictive region, rearange generators and load nodes to get as many as in synthetic configuration
    ref_prod = pd.DataFrame({'datetime':nrel_prod['datetime']})
    ref_load = pd.DataFrame({'datetime':nrel_prod['datetime']})

    regions = corresp_regions_gens['zone'].unique()
    for region_fictive in regions:
        ref_prod_ = pd.DataFrame({'datetime': nrel_prod.index})
        ref_load_ = pd.DataFrame({'datetime': nrel_load.index})

        # Production generators
        ref_gens = corresp_regions_gens[corresp_regions_gens['zone']==region_fictive]['name'].unique()
        nref = len(ref_gens)
        syn_gens = prods_charac[prods_charac['zone'] == region_fictive]['name'].unique()
        nsyn = len(syn_gens)


        if nref <= nsyn:
            for i, col in enumerate(syn_gens):
                if i < (nref-1):
                    ref_prod_[col] = nrel_prod[ref_gens[i]].astype(float)
                else:
                    ref_prod_[col] = nrel_prod[ref_gens[(nref-1)]].astype(float)/(nsyn-nref+1)
        else:
            for i, col in enumerate(syn_gens):
                if i < (nsyn-1):
                    ref_prod_[col] = nrel_prod[ref_gens[i]].astype(float)
                else:
                    ref_prod_[col] = nrel_prod[ref_gens[i:len(ref_gens)]].sum(axis = 1)
        ref_prod_.drop(columns = ['datetime'], inplace=True)
        ref_prod = pd.concat([ref_prod, ref_prod_], axis=1)

        # Loads nodes with same principle
        ref_gens = corresp_regions_loads[corresp_regions_loads['zone']==region_fictive]['name'].unique()
        nref = len(ref_gens)
        syn_gens = loads_charac[loads_charac['zone'] == region_fictive]['name'].unique()
        nsyn = len(syn_gens)

        if nref <= nsyn:
            for i, col in enumerate(syn_gens):
                if i < (nref-1):
                    ref_load_[col] = nrel_load[ref_gens[i]].astype(float)
                else:
                    ref_load_[col] = nrel_load[ref_gens[(nref-1)]].astype(float)/(nsyn-nref+1)
        else:
            for i, col in enumerate(syn_gens):
                if i < (nsyn - 1):
                    ref_load_[col] = nrel_load[ref_gens[i]].astype(float)
                else:
                    ref_load_[col] = nrel_load[ref_gens[i:len(ref_gens)]].sum(axis=1).astype(float)
        ref_load_.drop(columns=['datetime'], inplace = True)
        ref_load = pd.concat([ref_load, ref_load_], axis = 1)

    # Set datetime limits and resample like synthetic chronics
    ref_prod.set_index('datetime', inplace = True)
    ref_load.set_index('datetime', inplace=True)
    ref_prod = ref_prod.resample(timestep).first()
    ref_load = ref_load.resample(timestep).first()

    start_date = params['start_date']
    end_date = params['end_date']
    ref_prod.index = [ts.replace(year = year) for ts in ref_prod.index]
    ref_load.index = [ts.replace(year=year) for ts in ref_load.index]
    ref_prod = ref_prod[(ref_prod.index >= start_date) & (ref_prod.index < end_date)]
    ref_load = ref_load[(ref_load.index >= start_date) & (ref_load.index < end_date)]

    return ref_prod, ref_load


import datetime as dt
from functools import partial
import os
import re

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d

from ..config import DispatchConfigManager, LoadsConfigManager, ResConfigManager


def make_generation_input_output_directories(input_folder, case, year, output_folder):

    dispatch_input_folder_case = os.path.join(input_folder, case, 'dispatch')
    dispatch_input_folder = os.path.join(dispatch_input_folder_case, str(year))
    dispatch_output_folder = os.path.join(output_folder, str(year))

    os.makedirs(dispatch_input_folder_case, exist_ok=True)
    os.makedirs(dispatch_input_folder, exist_ok=True)
    os.makedirs(dispatch_output_folder, exist_ok=True)

    return dispatch_input_folder, dispatch_input_folder_case, dispatch_output_folder


def generate_coarse_noise(params, data_type):
    """
    This function generates a spatially and temporally correlated noise.
    Because it may take a lot of time to compute a correlated noise on
    a too fine mesh, we recommend to first compute a correlated signal
    on a coarse mesh, and then use the interpolation function.

    Input:
        params: (dict) Defines the mesh dimensions and
            precision. Also define the correlation scales

    Output:
        (np.array) 3D autocorrelated noise
    """

    # Get computation domain size
    Lx = params['Lx']
    Ly = params['Ly']
    T = params['T']

    # Get the decay parameter for each dimension
    dx_corr = params['dx_corr']
    dy_corr = params['dy_corr']
    dt_corr = params[data_type + '_corr']

    # Compute number of element in each dimension
    Nx_comp = int(Lx // dx_corr + 1)
    Ny_comp = int(Ly // dy_corr + 1)
    Nt_comp = int(T // dt_corr + 1)

    # Generate gaussian noise input·
    output = np.random.normal(0, 1, (Nx_comp, Ny_comp, Nt_comp))

    return output

def interpolate_noise(computation_noise, params, locations, time_scale):
    """
    This interpolates an autocarrelated noise mesh, to make it more granular.

    Input:
        computation_noise: (np.array) Autocorrelated signal computed on a coarse mesh
        params: (dict) Defines the mesh dimensions and
            precision. Also define the correlation scales
        locations: (dict) Defines the location of the points of interest in the domain

    Output:
        (dict of np.array) returns one time series per location mentioned in dict locations
    """

    # Get computation domain size
    Lx = params['Lx']
    Ly = params['Ly']
    T = params['T']

    # Get the decay parameter for each dimension
    dx_corr = params['dx_corr']
    dy_corr = params['dy_corr']
    dt_corr = time_scale

    # Compute number of element in each dimension
    Nx_comp = int(Lx // dx_corr + 1)
    Ny_comp = int(Ly // dy_corr + 1)
    Nt_comp = int(T // dt_corr + 1)

    # Get interpolation temporal mesh size
    dt = params['dt']

    # Compute number of element in each dimension
    Nt_inter = T // dt + 1

    # Get coordinates
    x, y = locations

    # Get coordinates of closest points in the coarse mesh
    x_minus = int(x // dx_corr)
    x_plus = int(x // dx_corr + 1)
    y_minus = int(y // dy_corr)
    y_plus = int(y // dy_corr + 1)

    # 1st step : spatial interpolation

    # Initialize output
    output = np.zeros(Nt_comp)

    # Initialize sum of distances
    dist_tot = 0

    # For every close point, add the corresponding time series, weighted by the inverse
    # of the distance between them
    for x_neighbor in [x_minus, x_plus]:
        for y_neighbor in [y_minus, y_plus]:
            dist = 1 / (np.sqrt((x - dx_corr * x_neighbor) ** 2 + (y - dy_corr * y_neighbor) ** 2) + 1)
            output += dist * computation_noise[x_neighbor, y_neighbor, :]
            dist_tot += dist
    output /= dist_tot

    # 2nd step : temporal quadratic interpolation
    t_comp = np.linspace(0, int(T), int(Nt_comp), endpoint=True)
    t_inter = np.linspace(0, int(T), int(Nt_inter), endpoint=True)
    if Nt_comp == 2:
        f2 = interp1d(t_comp, output, kind='linear')
    elif Nt_comp == 3:
        f2 = interp1d(t_comp, output, kind='quadratic')
    elif Nt_comp > 3:
        f2 = interp1d(t_comp, output, kind='cubic')

    if Nt_comp >= 2:
        output = f2(t_inter)

    return output

def natural_keys(text):
    return int([ c for c in re.split('(\d+)', text) ][1])


def time_parameters(weeks, start_date):
    result = dict()
    start_date = pd.to_datetime(start_date, format='%Y-%m-%d')
    result['weeks'] = weeks
    result['start_date'] = start_date
    result['year'] = start_date.year
    return result


def updated_time_parameters_with_timestep(time_parameters, timestep):
    time_parameters['end_date'] = time_parameters['start_date'] + dt.timedelta(
        days=7 * int(time_parameters['weeks'])) - dt.timedelta(minutes=timestep)
    time_parameters['T'] = int(
        pd.Timedelta(
            time_parameters['end_date'] - time_parameters['start_date']
        ).total_seconds() // 60
    )
    return time_parameters


def generate_seeds(n_seeds, seed_for_loads=None, seed_for_res=None, seed_for_disp=None):

    default_seed = np.random.randint(low=0, high=2 ** 31)
    if seed_for_loads is not None:
        np.random.seed(seed_for_loads)
    else:
        np.random.seed(default_seed)
    seeds_for_loads = [np.random.randint(low=0, high=2 ** 31) for _ in
                       range(n_seeds)]
    if seed_for_res is not None:
        np.random.seed(seed_for_res)
    else:
        np.random.seed(default_seed)
    seeds_for_res = [np.random.randint(low=0, high=2 ** 31) for _ in
                       range(n_seeds)]
    if seed_for_disp is not None:
        np.random.seed(seed_for_disp)
    else:
        np.random.seed(default_seed)
    seeds_for_disp = [np.random.randint(low=0, high=2 ** 31) for _ in
                       range(n_seeds)]

    return seeds_for_loads, seeds_for_res, seeds_for_disp


def read_all_configurations(weeks, start_date, case, input_folder, output_folder):
    time_params = time_parameters(weeks, start_date)
    year = time_params['year']

    print(f'output_folder: {output_folder}')

    load_config_manager = LoadsConfigManager(
        name="Loads Generation",
        root_directory=input_folder,
        input_directories=dict(case=case, patterns='patterns'),
        required_input_files=dict(case=['loads_charac.csv', 'params.json'],
                                  patterns=['load_weekly_pattern.csv']),
        output_directory=output_folder
    )
    load_config_manager.validate_configuration()

    params, loads_charac, load_weekly_pattern = load_config_manager.read_configuration()

    res_config_manager = ResConfigManager(
        name="Renewables Generation",
        root_directory=input_folder,
        input_directories=dict(case=case, patterns='patterns'),
        required_input_files=dict(case=['prods_charac.csv', 'params.json'],
                                  patterns=['solar_pattern.npy']),
        output_directory=output_folder
    )

    params, prods_charac, solar_pattern = res_config_manager.read_configuration()

    params.update(time_params)
    params = updated_time_parameters_with_timestep(params, params['dt'])

    dispath_config_manager = DispatchConfigManager(
        name="Dispatch",
        root_directory=input_folder,
        output_directory=output_folder,
        input_directories=dict(params=case),
        required_input_files=dict(params=['params_opf.json'])
    )
    dispath_config_manager.validate_configuration()
    params_opf = dispath_config_manager.read_configuration()

    return (year, params, loads_charac, prods_charac, load_weekly_pattern,
            solar_pattern, params_opf)


def folder_name_pattern(base_name, n_scenarios):
    padding_size = len(str(int(n_scenarios)))
    return name_pattern(base_name, padding_size)


def name_pattern(base_name, padding_size):
    prefix = str(base_name) + '_' + '{:0{width}d}'
    return partial(prefix.format, width=padding_size)


def warn_if_output_folder_not_empty(output_folder):
    if len(os.listdir(output_folder)) != 0:
        input(f'The output folder {output_folder} is not empty, '
              f'proceed with caution. '
              'Press Enter to continue or Crtl-C to exit...')



import os
import datetime as dt
import pandas as pd

def main(input_folder, output_folder, load, prod_solar, prod_wind, params, params_loss, write_results = True):
    """
    :param input_folder (str): input folder in which pattern folder can be found
    :param output_folder (str): output folder to write loss chronic
    :param load(pandas.DataFrame): generated load with Chronix2grid module (for other usage)
    :param prod_solar (pandas.DataFrame): generated solar production with Chronix2grid module (for other usage)
    :param prod_wind (pandas.DataFrame): generated wind production with Chronix2grid module (for other usage)
    :param params (dict): dictionary with parameters concerning chronic generations in general
    :param params_loss (dict): dictionary with parameters concerning loss computation
    :param write_results (bool): whether to serialize results or not in scenario_folder_path (True by default)
    :return: pandas.Series representing provided  loss chronic
    """

    loss_pattern_path = os.path.join(input_folder, 'patterns', params_loss["loss_pattern"])
    loss = generate_valid_loss(loss_pattern_path, params)
    if write_results:
        loss.to_csv(os.path.join(output_folder,'loss.csv.bz2'), sep = ';')
    return loss

def generate_valid_loss(loss_pattern_path, params):
    # It is assumed that provided loss_pattern contains the requested time period and time step

    # Reading and parsing dates
    dateparse = lambda x: dt.datetime.strptime(x, '%d/%m/%Y %H:%M')
    loss_pattern = pd.read_csv(loss_pattern_path, usecols=[0, 1],
                                parse_dates=[0], date_parser=dateparse, sep = ';')
    loss_pattern.set_index(loss_pattern.columns[0], inplace=True)

    # Extract subset of loss-pattern corresponding to the period studied
    loss_pattern.index = loss_pattern.index.map(lambda t: t.replace(year=params['year']))
    datetime_index = pd.date_range(
        start=params['start_date'],
        end=params['end_date'],
        freq=str(params['dt']) + 'min')
    loss_pattern = loss_pattern.loc[datetime_index]
    loss_pattern = loss_pattern.head(len(loss_pattern)-1)  # Last value is lonely for another day (same treatment as load and renewable)
    return pd.Series(loss_pattern[loss_pattern.columns[0]])



import json
import os

import numpy as np

import chronix2grid.constants as cst


def parse_seed_arg(seed, arg_name, default_seed):
    if seed is not None:
        try:
            casted_seed = int(seed)
        except TypeError:
            raise RuntimeError(f'The parameter {arg_name} must be an integer')
    else:
        casted_seed = default_seed
    return casted_seed


def generate_default_seed():
    return np.random.randint(low=0, high=2 ** 31)


def dump_seeds(output_directory, seeds, scenario_name=''):
    with open(os.path.join(output_directory, scenario_name+'_'+cst.SEEDS_FILE_NAME), 'w') as f:
        json.dump(seeds, f)


# Call generation scripts n_scenario times with dedicated random seeds
from chronix2grid.GeneratorBackend import GeneratorBackend


def main(generator: GeneratorBackend, case, n_scenarios, input_folder, output_folder, scen_names,
         time_params, mode='LRTK', scenario_id=None,
         seed_for_loads=None, seed_for_res=None, seed_for_disp=None):
    """
    Main function for chronics generation. It works with three steps: load generation, renewable generation (solar and wind) and then dispatch computation to get the whole energy mix

    Parameters
    ----------
    generator (GeneratorBackend): implementation class to do the different generation
    case (str): name of case to study (must be a folder within input_folder)
    n_scenarios (int): number of desired scenarios to generate for the same timescale
    params (dict): parameters of generation, as returned by function chronix2grid.generation.generate_chronics.read_configuration
    input_folder (str): path of folder containing inputs
    output_folder (str): path where outputs will be written (intermediate folder case/year/scenario will be used)
    prods_charac (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    loads_charac (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    lines (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    solar_pattern (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    load_weekly_pattern (pandas.DataFrame): as returned by function chronix2grid.generation.generate_chronics.read_configuration
    mode (str): options to launch certain parts of the generation process : L load R renewable T thermal

    Returns
    -------

    """
    return generator.run(case, n_scenarios, input_folder, output_folder, scen_names, time_params, mode, scenario_id, seed_for_loads, seed_for_res,
                         seed_for_disp)



import os
import argparse
import numpy as np
import pandas as pd
import seaborn as sns
from itertools import product
from pathlib import Path
from matplotlib import pyplot as plt
#from entropy import spectral_entropy

class HydroKPI:

    def __init__(self, kpi_databanks):
        self.kpi_databanks = kpi_databanks

    def __hydro_in_prices(self,
                          norm_mw,
                          upper_quantile,
                          lower_quantile,
                          above_norm_cap,
                          below_norm_cap):

        '''
        '''

        # Get number of gen units
        n_units = norm_mw.shape[1]

        # Get the price at upper/lower quantile
        eu_upper_quantile = self.prices.quantile(upper_quantile)
        eu_lower_quantile = self.prices.quantile(lower_quantile)

        # Test if units are above/below normalized capacity
        is_gens_above_cap =  norm_mw.ge(eu_upper_quantile, axis=1)
        is_gens_below_cap =  norm_mw.le(eu_lower_quantile, axis=1)

        # Test if prices are greater/lower than defined quantiles
        is_price_above_q = self.prices > eu_upper_quantile
        is_price_below_q = self.prices < eu_lower_quantile

        # Stacking price bool condition for all hydro units along
        # the columns and rename them as units names
        is_price_above_q_gens = is_price_above_q[['price'] * n_units]
        is_price_above_q_gens.columns = norm_mw.columns

        is_price_below_q_gens = is_price_below_q[['price'] * n_units]
        is_price_below_q_gens.columns = norm_mw.columns

        # Match occurence of price high and full gen disaptch
        high_price_kpi = 100 * (is_price_above_q_gens & is_gens_above_cap).sum(axis=0) \
                         / is_price_above_q_gens.sum(axis=0)

        # Match occurence when price is lower and almost no dispatch
        low_price_kpi = 100 * (is_price_below_q_gens & is_gens_below_cap).sum(axis=0) \
                        / is_price_below_q_gens.sum(axis=0)

        return high_price_kpi.round(self.precision), low_price_kpi.round(self.precision)

    def __hydro_seasonal(self, hydro_mw):

        '''
        '''

        # We aggregate hydro per month if a user want to deliver less MW in some
        # months rather than others.
        #
        # E.g. Typical configuration:
        # 6 month full capacity and
        # other months with 0 MW)
        hydro_mw_month = hydro_mw.copy()
        hydro_mw_month['month'] = self.months
        mw_per_month = hydro_mw_month.groupby('month').sum(axis=0).round(self.precision)

        return mw_per_month

    def hydro_kpi(self,
                  upper_quantile = 0.95,
                  lower_quantile = 0.05,
                  above_norm_cap = 0.9,
                  below_norm_cap = 0.1):

        '''
        Get 4 different Hydro KPI's based on the assumption the total costs
        of the system follow same curve as the consumption.

        Parameters:
        ----------

        upper_quantile (float): Quantile that define high prices.
                                (Prices are considered high whether
                                price(t) is greater than upper_quantile)
        lower_quantile (float): Quantile that define lower prices
                                (Prices are considered low whether
                                price(t) is less than lower_quantile)

        above_cap (float): Per unit (<1) criteria to establish high hydro dispatch
        below_cap (float): Per unit (<1) criteria to establish low hydro dispatch

        Returns:
        --------

        highPrice_kpi (dataframe): Percentage of time a generator is keeping
                                   operating above its predefined capacity

        lowPrice_kpi (dataframe): Percentage of time a generator is keeping
                                   operating below its predefined capacity

        mw_per_month (dataframe): Aggregated sum per month. Used to design
                                  seasonal pattern in hydro plants.
        '''

        # Get Hydro names
        hydro_filter = self.prod_charac.type.isin(['hydro'])
        hydro_names = self.prod_charac.name.loc[hydro_filter].values

        # Normalize MW according to the max value for
        # the reference data and synthetic one
        hydro_ref = self.ref_dispatch[hydro_names]
        hydro_syn = self.syn_dispatch[hydro_names]

        max_mw_ref = hydro_ref.max(axis=0)
        max_mw_syn = hydro_syn.max(axis=0)

        norm_mw_ref = hydro_ref / max_mw_ref
        norm_mw_syn = hydro_syn / max_mw_syn

        # Stats for reference data
        stat_ref_high_price, stat_ref_low_price = self.__hydro_in_prices(norm_mw_ref,
                                                                         upper_quantile,
                                                                         lower_quantile,
                                                                         above_norm_cap,
                                                                         below_norm_cap)

        # Stats for synthetic data
        stat_syn_high_price, stat_syn_low_price = self.__hydro_in_prices(norm_mw_syn,
                                                                         upper_quantile,
                                                                         lower_quantile,
                                                                         above_norm_cap,
                                                                         below_norm_cap)

        # Write results
        # -- + -- + --
        self.output['Hydro'] = {'high_price_for_ref': stat_ref_high_price.to_dict(),
                                'low_price_for_ref': stat_ref_low_price.to_dict(),
                                'high_price_for_syn': stat_syn_high_price.to_dict(),
                                'low_price_for_syn': stat_syn_low_price.to_dict()
                                }

        # Seasonal for reference data
        ref_agg_mw_per_month = self.__hydro_seasonal(hydro_ref)

        # Seasonal for synthetic data
        syn_agg_mw_per_month = self.__hydro_seasonal(hydro_syn)

        # Write results
        # -- + -- + --
        self.output['Hydro'] = {'seasonal_month_for_ref': ref_agg_mw_per_month.to_dict(),
                                'seasonal_month_for_syn': syn_agg_mw_per_month.to_dict()
                                }

        return stat_ref_high_price, stat_ref_low_price, ref_agg_mw_per_month, \
               stat_syn_high_price, stat_syn_low_price, syn_agg_mw_per_month



from abc import ABC, abstractmethod
import json
import os

import numpy as np
import pandas as pd


class ConfigManager(ABC):
    """
    Class that validates a configuration setting by checking that the input files are correctly provided.
    It also provided a static method for specific reading of different input files

    Attributes
    ----------
    name: ``str``
        Name of config manager
    root_directory: ``str``
    input_directories: ``str`` or ``dict``
    output_directory: ``str``
    required_input_files: ``list`` or ``None``
        dict with compulsory files to be checked
    """
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        """
        Self initialization

        Parameters
        ----------
        name: ``str``
        root_directory: ``str``
        input_directories: ``str`` or ``dict``
        output_directory: ``str``
        required_input_files: ``str``
        required_input_files: ``dict`` or ``None``
        """
        self.name = name
        self.root_directory = root_directory
        self.input_directories = input_directories
        self.output_directory = output_directory
        self.required_input_files = required_input_files if required_input_files is not None else []

    def is_single_input_dir(self):
        if isinstance(self.input_directories, str):
            return True
        if isinstance(self.input_directories, dict):
            return False
        raise RuntimeError("input_directories must be either a string or a dictionnary")

    def validate_input(self):
        """
        Validate that the input folder exists and contains expected files. Returns an error message with :func:`ConfigManager.error_message`
        if not the case
        """
        if self.is_single_input_dir():
            directory_to_check = os.path.join(self.root_directory, self.input_directories)
            try:
                files_to_check = os.listdir(directory_to_check)
                for config_file in self.required_input_files:
                    if config_file not in files_to_check:
                        return False
            except FileNotFoundError:
                raise FileNotFoundError(self.error_message())
        else:
            for input_name, input_path in self.input_directories.items():
                input_directory_abs_path = os.path.join(self.root_directory, input_path)
                try:
                    files_to_check = os.listdir(input_directory_abs_path)
                    for config_file in self.required_input_files[input_name]:
                        if config_file not in files_to_check:
                            return False
                except FileNotFoundError:
                    raise FileNotFoundError(self.error_message())
        return True

    def validate_output(self):
        """
        Check if the output path exists from root directory
        """
        return os.path.exists(os.path.join(self.root_directory, self.output_directory))

    def error_message(self):
        """
        Returns an error message if inputs and outputs are not valid

        """
        output_directory_abs_path = os.path.join(
            self.root_directory, self.output_directory)
        error_message_header = (
            f"\n{self.name} process requires the following configuration "
            f"(directories and files must exist): \n"
            f"  - Root directory is {self.root_directory}\n"
            f"  - Output directory is {output_directory_abs_path}\n"
        )
        if self.is_single_input_dir():
            input_directory_abs_path = os.path.join(
                self.root_directory, self.input_directories)
            formatted_required_input_files = ""
            for filename in sorted(self.required_input_files):
                formatted_required_input_files += "\n      - " + filename

            error_msg_body = (
                    f"  - Input directory is {input_directory_abs_path}\n"
                    f"    - Expected input files are:" + formatted_required_input_files
            )
        else:
            formatted_required_directories = ""
            for input_name, input_path in self.input_directories.items():
                input_directory_abs_path = os.path.join(self.root_directory, input_path)
                formatted_required_directories += (
                        "\n    - " + input_directory_abs_path +
                        "\n      - with expected input files:"
                )
                for required_file in self.required_input_files[input_name]:
                    formatted_required_directories += "\n        - " + required_file

            error_msg_body = (
                    f"  - Input directories are:" + formatted_required_directories
            )
        return error_message_header + error_msg_body

    def validate_configuration(self):
        if not self.validate_output() or not self.validate_input():
            raise FileNotFoundError(self.error_message())
        return True

    @abstractmethod
    def read_configuration(self):
        pass

class GeneralConfigManager(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(GeneralConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        """
        Reads parameters for the overall generation process

            * *dt* - temporal granularity of the process
            * *planned_std* - noise level for forecast chronics

        Returns
        -------
        params: ``dict``
            dictionary of parameters
        """
        params_file_path = os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params.json')
        with open(params_file_path, 'r') as json1_file:
            json1_str = json1_file.read()
        params = json.loads(json1_str)
        for key, value in params.items():
            try:
                params[key] = float(value)
            except ValueError:
                params[key] = pd.to_datetime(value, format='%Y-%m-%d')
        return params

class LoadsConfigManager(ConfigManager):
    """
    Reads parameters for :class:`chronix2grid.generation.consumption.ConsumptionGeneratorBackend`

        * *Lx* - x dimension of coarse grid for spatially correlated noise
        * *Ly* - y dimension of coarse grid for spatially correlated noise
        * *dx_corr* - x granularity of coarse grid for spatially correlated noise
        * *dy_corr* - y granularity of coarse grid for spatially correlated noise
        * *temperature_corr* - noise level for spatially correlated noise
        * *std_temperature_noise* - noise level for temporally autocorrelated noise

    Returns
    -------
    params_load: ``dict``
        dictionary of parameters
    """
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(LoadsConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        params_file_path = os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params_load.json')
        with open(params_file_path, 'r') as json1_file:
            json1_str = json1_file.read()
        params = json.loads(json1_str)
        for key, value in params.items():
            try:
                params[key] = float(value)
            except ValueError:
                params[key] = pd.to_datetime(value, format='%Y-%m-%d')

        # Nt_inter = int(params['T'] // params['dt'] + 1)
        try:
            loads_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'loads_charac.csv'),
                sep=',')
            names = loads_charac['name']  # to generate error if separator is wrong

        except:
            loads_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'loads_charac.csv'),
                sep=';')

        return params, loads_charac

    def read_specific(self):
        """
        Reads data frame with loads characteristics

        Returns
        -------
        loads_charac: :class:`pandas.DataFrame`
        """
        load_weekly_pattern = pd.read_csv(
            os.path.join(self.root_directory, self.input_directories['patterns'],
                         'load_weekly_pattern.csv'))
        return load_weekly_pattern

class ResConfigManager(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(ResConfigManager, self).__init__(name, root_directory, input_directories,
                                               output_directory, required_input_files)

    def read_configuration(self):
        """
        Reads parameters for :class:`chronix2grid.generation.renewable.RenewableBackend`

            * *Lx* - x dimension of coarse grid for spatially correlated noise
            * *Ly* - y dimension of coarse grid for spatially correlated noise
            * *dx_corr* - x granularity of coarse grid for spatially correlated noise
            * *dy_corr* - y granularity of coarse grid for spatially correlated noise
            * *solar_corr*, *short_wind_corr*, *medium_wind_corr*, *long_wind_corr* - noise levels for spatially correlated noises
            * *std_solar_noise*, *std_short_wind_noise*, *std_medium_wind_noise*, *std_long_wind_noise* - noise levels for temporally autocorrelated noises
            * *smoothdist* - independent noise level
            * *year_solar_pattern* - year of provided solar pattern

        Returns
        -------
        params_res: ``dict``
            dictionary of parameters
        """
        params_file_path = os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params_res.json')
        with open(params_file_path, 'r') as params_json:
            params = json.load(params_json)
        for key, value in params.items():
            try:
                params[key] = float(value)
            except ValueError:
                params[key] = pd.to_datetime(value, format='%Y-%m-%d')

        # Nt_inter = int(params['T'] // params['dt'] + 1)
        try:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=',')
            names = prods_charac['name']  # to generate error if separator is wrong

        except:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=';')

        return params, prods_charac

    def read_specific(self):
        """
        Reads data frame with generator characteristics

        Returns
        -------
        prods_charac: :class:`pandas.DataFrame`
        """
        solar_pattern = np.load(
            os.path.join(self.root_directory, self.input_directories['patterns'],
                         'solar_pattern.npy'))
        return solar_pattern

class ResConfigManagerGan(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(ResConfigManagerGan, self).__init__(name, root_directory, input_directories,
                                               output_directory, required_input_files)

    def read_configuration(self):
        """
        Reads parameters for :class:`chronix2grid.generation.renewable.RenewableBackendGAN`

            * *model_name_wind* - name of tensorflow serialized model for wind
            * *batch_size_wind*, *n_gens_wind*, *n_timesteps_wind* - dimensions of learning batches of wind networks
            * *n_events_wind* -  number of labels handled as input, representing arbitrary events in learning of wind
            * *dim_input_wind*, *mu_wind*, *sigma_wind* - characteristics of inpus (size of gaussian noise vector, mean and standard deviation)

        ALl the same parameters are expected for solar generation (with suffix *_solar*), that will use the same network architecture but with potentialy différent layout parameters

        Returns
        -------
        params_res: ``dict``
            dictionary of parameters
        prods_charac: :class:`pandas.DataFrame`
        """
        params_file_path = os.path.join(
            self.root_directory,
            self.input_directories['case'], 'params.json')
        network_folder = os.path.join(self.root_directory,
                                      self.input_directories['case'],
                                      'neural_network')
        params_file_path_gan = os.path.join(network_folder, 'paramsGAN.json')

        # Check timestep
        with open(params_file_path, 'r') as params_json:
            params_general = json.load(params_json)
        if params_general["dt"] % 60 != 0:
            raise ValueError('timesteps different from 60min or multiples of 60min are not supported yet with GAN. Please use another model')

        with open(params_file_path_gan, 'r') as params_json:
            params = json.load(params_json)

        for key, value in params.items():
            if key in ["mu_wind","sigma_wind","mu_solar","sigma_solar"]:
                params[key] = float(value)
            elif "model_name" not in key:
                try:
                    params[key] = int(value)
                except ValueError:
                    params[key] = pd.to_datetime(value, format='%Y-%m-%d')

        try:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=',')
            names = prods_charac['name']  # to generate error if separator is wrong

        except:
            prods_charac = pd.read_csv(
                os.path.join(self.root_directory, self.input_directories['case'],
                             'prods_charac.csv'),
                sep=';')
        params["network_folder"] = network_folder

        return params, prods_charac

class DispatchConfigManager(ConfigManager):
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(DispatchConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        """
        Reads parameters for :class:`chronix2grid.generation.dispatch.DispatchBackend`

            * *step_opf_min* - time resolution of the OPF in minutes. It can be 5, 10, 15, 20, 30 or 60 and has to be superior or equal to dt (generation time resolution). In case it is strictly above, interpolation is done after dispatch resolution
            * *mode_opf* - frequency at which we wan't to solve the OPF
            * *dispatch_by_carrier* - if True, dispatch results will be returned for the whole carrier. If False, it will be returned by generator
            * *ramp_mode* is essentially designed for debug purpose: when your OPF diverges, you may want to relax some constraints to know the reasons why the problem is unfeasible or leads to divergence
                * If *hard*, all the ramp constraints will be taken into account.
                * If *medium*, thermal ramp-constraints are skipped
                * If *easy*, thermal and hydro ramp-constraints are skipped
                * If *none*, thermal, hydro and nuclear ramp-constraints are skipped
            * *reactive_comp* - Factor applied to consumption to compensate reactive part not modelled by linear opf
            * *pyomo* - whether pypsa should use pyomo or not (boolean)
            * *solver_name* - name of solver, that you should have installed in your environment and added in your environment variables.
            * *hydro_ramp_reduction_factor* - optional factor which will divide max ramp up and down to all hydro generators
            * *losses_pct**- if D mode is deactivate, losses are estimated as a percentage of load.

        Optional parameters can be set for grid2op simulation of loss as a final step.
        The production is updated on a slack generator and warnings or errors are returned if this update violates generator constraints

        * *slack_p_max_reduction* - before dispatch, reduce Pmax of slack generator temporary to anticipate loss
        * *slack_ramp_max_reduction* - before dispatch, reduce ramp max (up and down) of slack generator temporary to anticipate loss
        * *loss_grid2op_simulation* - if True, launches grid2Op simulation for loss
        * *idxSlack*, *nameSlack* - identifies slack generator that will be updated
        * *early_stopping_mode* if True returns errors if generator constraints are violated after updates. If False, only returns warnings
        * *agent_type* - Grid2op agent type ti use for simulation. Can be "reco" for RecoPowerLines or "do-nothing"

        .. warning::
            The dispatch optimization can rely on pypsa simulation. If it is the case you should ensure pypsa dependencies are installed

        .. note::
            If no *loss_grid2op_simulation* is provided, chronix2grid follows considering it is False

        Returns
        -------
        params_opf: ``dict``
            dictionary of parameters
        """

        # Basics
        self.validate_configuration()
        params_filepath = os.path.join(
            self.root_directory,
            self.input_directories['params'],
            'params_opf.json')
        with open(params_filepath, 'r') as opf_param_json:
            params_opf = json.load(opf_param_json)
        try:
            if params_opf['mode_opf'] == '':
                params_opf['mode_opf'] = None
        except KeyError:
            raise KeyError('The mode_opf field of params_opf.json is missing.')

        # Grid2op loss simulation (optional)
        bool = False
        try:
            bool = params_opf["loss_grid2op_simulation"]
        except:
            print('Warning: The loss_grid2op_simulation field of params_opf.json is missing. Continuing assuming it is False')
            params_opf["loss_grid2op_simulation"] = False
        if bool:
            oblig_keys = ["idxSlack","nameSlack","agent_type"]
            for key in oblig_keys:
                try:
                    params_opf[key]
                except KeyError:
                    raise KeyError('loss_grid2op_simulation is set to True, key '+str(key) + ' must be provided')

        # Hydro correction
        if "hydro_ramp_reduction_factor" not in list(params_opf.keys()):
            params_opf["hydro_ramp_reduction_factor"] = 1.
        else:
            params_opf["hydro_ramp_reduction_factor"] = float(params_opf["hydro_ramp_reduction_factor"])

        # Slack temporary correction
        for key in ["slack_p_max_reduction", "slack_ramp_max_reduction"]:
            if key not in list(params_opf.keys()):
                params_opf[key] = 0.
            else:
                params_opf[key] = float(params_opf[key])
        return params_opf


def read_all_configuration(files, root_directory, input_directories):
    params = {}
    for file in files:
        params_file_path = os.path.join(
            root_directory,
            input_directories['case'], file)
        with open(params_file_path, 'r') as params_json:
            params_to_add = json.load(params_json)
        for key, value in params_to_add.items():
            try:
                params_to_add[key] = float(value)
            except ValueError:
                params_to_add[key] = pd.to_datetime(value, format='%Y-%m-%d')
        params.update(params_to_add)
    return params

class LossConfigManager(ConfigManager):
    """
            Checks parameters for :class:`chronix2grid.generation.loss.LossBackend`
                * *loss_pattern* - name of loss pattern to read

            Returns
            -------
            params_loss: ``dict``
                dictionary of parameters
            """
    def __init__(self, name, root_directory, input_directories, output_directory,
                 required_input_files=None):
        super(LossConfigManager, self).__init__(name, root_directory, input_directories,
                                                 output_directory, required_input_files)

    def read_configuration(self):
        self.validate_configuration()
        params_filepath = os.path.join(
            self.root_directory,
            self.input_directories['params'],
            'params_loss.json')
        with open(params_filepath, 'r') as loss_param_json:
            params_loss = json.load(loss_param_json)
        try:
            if params_loss['loss_pattern'] == '':
                params_loss['loss_pattern'] = 'loss_pattern.csv'
        except KeyError:
            raise KeyError('The loss_pattern field of params_loss.json is missing.')
        return params_loss



import seaborn as sns


def _plot_heatmap(self, corr, title, path_png=None, save_png=False):

    ax = sns.heatmap(corr,
                     fmt='.1f',
                     annot = False,
                     vmin=-1, vmax=1, center=0,
                     cmap=sns.diverging_palette(20, 220, n=200),
                     cbar_kws={"orientation": "horizontal",
                               "shrink": 0.35,
                               },
                     square=True,
                     linewidths = 0.5,
                     )
    ax.set_xticklabels(ax.get_xticklabels(),
                       rotation=45,
                       horizontalalignment='right'
                       )

    ax.set_title(title, fontsize=15)

    if save_png:
        figure = ax.get_figure()
        figure.savefig(path_png)



"""
This file gathers constants related to directories and files that chronix2grid
reads/writes. It also defines the ConfigManager and the Backend used for each step of generation process (LRDT)

.. note::
    For the moment you can only change the backend for renewable generation and the associated config manager.

    Switch from RenewableBackend to RenewableBackendGAN and from ResConfigManager to ResConfigManagerGan

    Note that this would require tensorflow 1.15.4 which is an optional dependency of chronix2grid
"""

##############################################
################## NAMES #####################
##############################################

GENERATION_FOLDER_NAME = 'generation'
KPI_FOLDER_NAME = 'kpi'
KPI_IMAGES_FOLDER_NAME = 'images'

DEFAULT_OUTPUT_FOLDER_NAME = 'output'
DEFAULT_INPUT_FOLDER_NAME = 'getting_started/example/input'

SCENARIO_FOLDER_BASE_NAME = 'Scenario'

SEEDS_FILE_NAME = 'seeds_info.json'

FLOATING_POINT_PRECISION_FORMAT = '%.1f'

TIME_STEP_FILE_NAME = 'time_interval.info'

REFERENCE_ZONE = 'France'

GRID_FILENAME = 'grid.json'

##############################################
################ BACKENDS ####################
##############################################

from chronix2grid.generation.consumption.ConsumptionGeneratorBackend import ConsumptionGeneratorBackend
from chronix2grid.generation.renewable.RenewableBackend import RenewableBackend
from chronix2grid.generation.loss.LossBackend import LossBackend
from chronix2grid.generation.dispatch.DispatchBackend import DispatchBackend
from chronix2grid.config import GeneralConfigManager, LoadsConfigManager, ResConfigManager, LossConfigManager, DispatchConfigManager

GENERAL_CONFIG = GeneralConfigManager

#### LOAD (L) ####
LOAD_GENERATION_CONFIG = LoadsConfigManager
LOAD_GENERATION_BACKEND = ConsumptionGeneratorBackend

#### RENEWABLE - SOLAR AND WIND (R) ####
# from chronix2grid.config import ResConfigManagerGan
RENEWABLE_GENERATION_CONFIG = ResConfigManager #ResConfigManagerGan #ResConfigManager
# from chronix2grid.generation.renewable.RenewableBackend import RenewableBackendGAN
RENEWABLE_GENERATION_BACKEND = RenewableBackend #RenewableBackendGAN # RenewableBackend

#### LOSS (D) ####
LOSS_GENERATION_CONFIG = LossConfigManager
LOSS_GENERATION_BACKEND = LossBackend

#### DISPATCH - HYDRO, THERMAL, NUCLEAR (T) ####
DISPATCH_GENERATION_CONFIG = DispatchConfigManager
HYDRO_GENERATION_BACKEND = None
from PypsaDispatchBackend.PypsaEconomicDispatch import PypsaDispatcher
DISPATCHER = PypsaDispatcher
DISPATCH_GENERATION_BACKEND = DispatchBackend

#### KPI (K) ####
RENEWABLE_NINJA_REFERENCE_FOLDER = 'renewable_ninja'
GAN_TRAINING_SET_REFERENCE_FOLDER = 'GAN_training_data'


import os
from enum import Enum
import copy

import numpy as np

class RampMode(Enum):
    """
    Encodes the level of complexity of the ramp constraints to apply for
    the economic dispatch
    """
    none = -1
    easy = 0
    medium = 1
    hard = 2


def make_scenario_input_output_directories(input_folder, output_folder, scenario_name):
    os.makedirs(os.path.join(input_folder, scenario_name), exist_ok=True)
    os.makedirs(os.path.join(output_folder, scenario_name), exist_ok=True)
    return os.path.join(input_folder, scenario_name), os.path.join(output_folder, scenario_name)

def modify_hydro_ramps(grid2op_env, hydro_dividing_factor=1., decimals=1):
    for i, generator in enumerate(grid2op_env.name_gen):
        gen_type = grid2op_env.gen_type[i]
        if gen_type == "hydro":
           grid2op_env.gen_max_ramp_up[i] = np.round(grid2op_env.gen_max_ramp_up[i]/hydro_dividing_factor, decimals)
           grid2op_env.gen_max_ramp_down[i] = np.round(grid2op_env.gen_max_ramp_down[i]/hydro_dividing_factor, decimals)
    return grid2op_env

def modify_slack_characs(grid2op_env, slack_name, p_max_reduction=0, ramp_reduction=0, decimals = 1):
    for i, generator in enumerate(grid2op_env.name_gen):
        if generator == slack_name:
            grid2op_env.gen_pmax[i] = np.round(grid2op_env.gen_pmax[i] - p_max_reduction, decimals)
            grid2op_env.gen_max_ramp_up[i] = np.round(grid2op_env.gen_max_ramp_up[i] - ramp_reduction, decimals)
            grid2op_env.gen_max_ramp_down[i] = np.round(grid2op_env.gen_max_ramp_down[i] - ramp_reduction, decimals)
    return grid2op_env

def add_noise_gen(dispatch, gen_cap, noise_factor):
    """ Add noise to opf dispatch to have more
    realistic real-time data

    Parameters
    ----------
    dispatch : dataframe
        Opf PyPSA output
    gen_cap : dataframe
        Maximun capacity for gen
    noise_factor : float
        Noise factor applied to every gen col

    Returns
    -------
    dataframe
        Distpach with noise
    """
    dispatch_new = copy.deepcopy(dispatch)  # dispatch.copy(deep=True)

    # variance_per_col = gen_cap * noise_factor
    print('applying noise to forecast of ' + str(noise_factor) + ' %')
    for col in list(dispatch_new):
        # Check for values greater than zero
        # (means unit has been distpached)
        # only_dispatched_steps = dispatch_new[col][dispatch_new[col] > 0]
        # print(only_dispatched_steps)

        noise = np.random.lognormal(mean=0.0, sigma=noise_factor, size=dispatch_new.shape[0])
        dispatch_new[col] = dispatch[col] * noise
    return dispatch_new.round(2)


import os
import warnings
import shutil
import numpy as np
import pandas as pd

import grid2op
from grid2op.Chronics import Multifolder, GridStateFromFileWithForecasts
from grid2op.Parameters import Parameters
from grid2op.Runner import Runner
from grid2op.Episode import EpisodeData

#from grid2op.Agent import RecoPowerlineAgent
from grid2op.Agent import DoNothingAgent

import chronix2grid.constants as cst

def move_chronics_temporarily(scenario_output_folder, grid_path):
    chronics_temporary_path = os.path.join(grid_path,'chronics')
    if os.path.exists(chronics_temporary_path):
        warnings.warn("Had to delete a previous chronic temporary path in input data", UserWarning)
        shutil.rmtree(chronics_temporary_path)
    os.makedirs(chronics_temporary_path, exist_ok=True)
    chronics_temporary_path = os.path.join(chronics_temporary_path, '000')
    print("temporary copy of chronics in "+str(chronics_temporary_path))
    shutil.copytree(scenario_output_folder, chronics_temporary_path)

def remove_temporary_chronics(grid_path):
    chronics_temporary_path = os.path.join(grid_path, 'chronics')
    shutil.rmtree(chronics_temporary_path)

def remove_simulation_data(scenario_folder_path):
    simulation_data_folder = os.path.join(scenario_folder_path, 'loss_simulation')
    shutil.rmtree(simulation_data_folder)

def create_or_replace_simulation_data_folder(scenario_folder_path):
    simulation_data_folder = os.path.join(scenario_folder_path, 'loss_simulation')
    if os.path.exists(simulation_data_folder):
        shutil.rmtree(simulation_data_folder)
    os.makedirs(simulation_data_folder, exist_ok=False)
    return simulation_data_folder

def run_grid2op_simulation_donothing(grid_path, agent_result_path, agent_type = 'do-nothing', nb_core = 1):
    """

    :param grid_path (str): path to folder where grid.json and other information on grid are stored
    :param agent_result_path (str): path in which we want to write the result of Grid2op simulation (via serialized EpisodeData object)
    :param agent_type (str): type of grid2op agent to use in simulation. Can be reco for RecoPowerlineAgent or do-nothing for DoNothingAgent
    :param nb_core (int): Simulations can be paralelized on cores. By default, 1 core is used here because paralelization is made previously on chronic scenarios
    :return:
    """
    print('Start grid2op simulation to compute realistic loss on grid')

    # Récupération des paramètres
    NB_CORE = nb_core # TOUJOURS 1 car la parallélisation se fait sur les scénarii dans le main
    nb_episode = 1
    simulation_data_folder = create_or_replace_simulation_data_folder(agent_result_path)

    print('Grid2op simulation for loss')
    print('the case folder is: ' + grid_path)
    print('the output folder will be: ' + simulation_data_folder)
    print('the number of cores used is: ' + str(NB_CORE))
    print('the number of scenarios we consider is: ' + str(nb_episode))
    try:
        from lightsim2grid.LightSimBackend import LightSimBackend
        backend = LightSimBackend()
    except:
        from grid2op.Backend import PandaPowerBackend
        backend = PandaPowerBackend()
        print("You might need to install the LightSimBackend (provisory name) to gain massive speed up")
    # don't disconnect powerline on overflow, the thermal limit are not set for now, it would not make sens
    param = Parameters()
    param.init_from_dict({"NO_OVERFLOW_DISCONNECTION": True})

    env = grid2op.make(grid_path,
                       param=param, backend=backend, test=True)
    #env = grid2op.make(grid_path = grid_path, chronics_path = scenario_output_folder,
     #                  param=param, backend=backend, test=True)
    # If you remove the "GridStateFromFileWithForecasts", from above, chronics will NOT be loaded properly.
    # GridStateFromFileWithForecasts is the format used for the competition, so it is mandatory that this works!
    # WITHOUT ANY MODIFICATIONS

    # Beside the environment should be able to load all data generated, and not one episode.
    # so please look in grid2op for compatible formats. This is not a valid format.

    if (agent_type == 'reco'):
        runner = Runner(**env.get_params_for_runner(), agentClass=DoNothingAgent)
        #runner = Runner(**env.get_params_for_runner(), agentClass=RecoPowerlineAgent)
    else:
        runner = Runner(**env.get_params_for_runner())
    # do regular computation as you would with grid2op
    res = runner.run(nb_episode=nb_episode, nb_process=NB_CORE, path_save=simulation_data_folder, pbar=True)
    print('---- end of simulation')

def correct_scenario_loss(scenario_folder_path, agent_result_path, params_opf, grid_path):
    print('Start realistic loss correction from simulation results')

    # Load simulation data
    data_this_episode = EpisodeData.from_disk(os.path.join(agent_result_path, 'loss_simulation'), '000')
    slack_name = params_opf["nameSlack"]
    id_slack = params_opf["idxSlack"]

    # Get gen constraints
    first_obs = data_this_episode.observations[0]
    pmax = first_obs.gen_pmax[id_slack] #+ params_opf['pmax_margin']
    pmin = first_obs.gen_pmin[id_slack] #max(first_obs.gen_pmin[id_slack] - params_opf['pmin_margin'],0)
    ramp_up = first_obs.gen_max_ramp_up[id_slack] #+ params_opf['rampup_margin']
    ramp_down = first_obs.gen_max_ramp_down[id_slack] #+ params_opf['rampdown_margin']

    # Get corrected dispatch prod
    prods_p = pd.DataFrame(np.array([obs.prod_p for obs in data_this_episode.observations]))
    prodSlack = prods_p[id_slack]

    # Get dispatch prods before runner in chronix
    OldProdsDf = pd.read_csv(os.path.join(scenario_folder_path, 'prod_p.csv.bz2'), sep=';')
    OldProdsForecastDf = pd.read_csv(os.path.join(scenario_folder_path, 'prod_p_forecasted.csv.bz2'), sep=';')

    ##correction term
    newProdsDf = OldProdsDf
    newProdsForecastDf = OldProdsForecastDf
    CorrectionLosses = prodSlack - OldProdsDf[slack_name]
    print('maximum compensation for slack before correction in MW: ' + str(CorrectionLosses.abs().max()))
    print('average compensation for slack before correction in MW: ' + str(CorrectionLosses.mean()))
    print('median compensation for slack before correction in MW: ' + str(CorrectionLosses.abs().median()))
    print('min compensation for slack before correction in MW: ' + str(CorrectionLosses.min()))

    # apply correction on slack bus generator
    newProdsDf[slack_name] = OldProdsDf[slack_name] + CorrectionLosses
    newProdsForecastDf[slack_name] = OldProdsForecastDf[slack_name] + CorrectionLosses

    # Check constraints
    violations_message, bool = check_slack_constraints(newProdsDf[slack_name], pmax, pmin, ramp_up, ramp_down)
    if bool:
        if params_opf['early_stopping_mode']:
            remove_temporary_chronics(grid_path)
            raise ValueError(violations_message)
        else:
            warnings.warn(violations_message, UserWarning)
            print("Warning - "+violations_message)


    # Serialization
    newProdsDf.to_csv(
            os.path.join(scenario_folder_path, "prod_p.csv.bz2"),
            sep=';', index=False,
            float_format=cst.FLOATING_POINT_PRECISION_FORMAT
        )

    newProdsForecastDf.to_csv(
        os.path.join(scenario_folder_path, "prod_p_forecasted.csv.bz2"),
        sep=';', index=False,
        float_format=cst.FLOATING_POINT_PRECISION_FORMAT
    )

    print('---- end of loss correction ')
    return newProdsDf, newProdsForecastDf

def check_slack_constraints(prod_p, pmax, pmin, ramp_up, ramp_down):
    msg = "Loss correction violates generator constraints: "
    bool = False

    # Pmax
    dep = max(prod_p-pmax)
    if dep > 0:
        bool = True
        msg += "Pmax + margin is violated with maximum of "+str(dep)+" MW - "

    # Pmin
    dep = min(prod_p-pmin)
    if dep < 0:
        bool = True
        msg += "Pmin - margin is violated with maximum of " + str(dep) + " MW - "

    # Ramp up
    ramps = prod_p.diff()
    ramps_up = ramps[ramps>0]
    dep = max(ramps_up - ramp_up)
    if dep > 0:
        bool = True
        msg += "Ramp up + margin is violated with maximum of " + str(dep) + " MW - "

    # Ramp down
    ramps_down = -1 * ramps[ramps < 0]
    dep = max(ramps_down - ramp_down)
    if dep > 0:
        bool = True
        msg += "Ramp down + margin is violated with maximum of " + str(dep) + " MW - "

    return msg, bool


import os
import tensorflow as tf
import numpy as np
import pandas as pd

def load_model(sess, params, network_folder, carrier):

    saver = tf.compat.v1.train.import_meta_graph(os.path.join(network_folder, carrier, params["model_name_"+carrier]+".meta"))
    checkpoint = tf.compat.v1.train.latest_checkpoint(os.path.join(network_folder, carrier))
    saver.restore(sess, checkpoint)
    dcgan_model = ReplayedGAN(
        dim_y=params["n_events_"+carrier],
        batch_size=params["batch_size_"+carrier],
        image_shape=[params["n_timesteps_"+carrier],params["n_gens_"+carrier],1],
        dim_z=params["dim_inputs_"+carrier]
    )
    del saver
    print("Replayed_W_DCGAN model loaded for "+carrier)
    return dcgan_model, sess

def OneHot(X, n, negative_class=0.):
    X = np.asarray(X).flatten()
    if n is None:
        n = np.max(X) + 1
    Xoh = np.ones((len(X), n)) * negative_class
    for i in range(len(X)):
        m=X[i]
        Xoh[i,m]=1
    return Xoh

def generate_gaussian_inputs(params, carrier):
    n_preds = compute_n_preds(params, carrier)
    Y = []
    Z = []
    for pred in range(n_preds):
        Y.append(OneHot(np.random.randint(int(params["n_events_"+carrier]), size=[params["batch_size_"+carrier]]), n=int(params["n_events_"+carrier])))
        Z.append(np.random.normal(params["mu_"+carrier], params["sigma_"+carrier], size=[int(params["batch_size_"+carrier]), int(params["dim_inputs_"+carrier])]).astype(np.float32))
    return Y,Z

def compute_n_preds(params, carrier):
    n_days_by_pred = params["batch_size_"+carrier]
    n_days = (int(params["T"] / params["dt"]) // 24) + 1
    n_preds = (n_days // n_days_by_pred) + 1
    return n_preds

def post_process_sample(generated_batches, params, prods_charac, datetime_index, carrier = "wind"):
    gens = prods_charac[prods_charac['type']==carrier]["name"].unique()
    wind = pd.DataFrame(columns = gens)
    N = len(gens)
    N_learning = params['n_gens_'+carrier]

    if N > N_learning:
        raise ValueError("the neural network should be trained on at least the same number of generators as in the generation process")
    else:
        if N%2==0:
            selection_gens = [i for i in range(N//2)] + [i for i in range((N_learning-N//2), N_learning)]
        else:
            selection_gens = [i for i in range((N // 2)+1)] + [i for i in range((N_learning - N // 2), N_learning)]

    for day, batch in enumerate(generated_batches):
        for i in range(params["batch_size_"+carrier]):
            matrix = batch[i,:,selection_gens,0] # batch[i,:,:len(gens),0] # DIMENSIONS
            df = pd.DataFrame(np.transpose(matrix), columns=gens) # Enlever ou ajouter np.transpose en conséquence matric --> np.transpose(matrix)
            wind = pd.concat([wind, df], axis = 0)
    # Truncate last batch
    wind = wind.reset_index(drop=True)
    wind = wind.iloc[:len(datetime_index),:]

    # Time index
    wind['datetime'] = datetime_index

    # Power rescaling
    for gen in gens:
        Pmax = prods_charac.loc[prods_charac['name']==gen,'Pmax'].values[0]
        wind[gen] = wind[gen] * Pmax
    return wind

def batchnormalize(X, eps=1e-8, g=None, b=None):
    if X.get_shape().ndims == 4:
        mean = tf.reduce_mean(X, [0,1,2])
        std = tf.reduce_mean( tf.square(X-mean), [0,1,2] )
        X = (X-mean) / tf.sqrt(std+eps)

        if g is not None and b is not None:
            g = tf.reshape(g, [1,1,1,-1])
            b = tf.reshape(b, [1,1,1,-1])
            X = X*g + b

    elif X.get_shape().ndims == 2:
        mean = tf.reduce_mean(X, 0)
        std = tf.reduce_mean(tf.square(X-mean), 0)
        X = (X-mean) / tf.sqrt(std+eps)

        if g is not None and b is not None:
            g = tf.reshape(g, [1,-1])
            b = tf.reshape(b, [1,-1])
            X = X*g + b

    else:
        raise NotImplementedError

    return X

def lrelu(X, leak=0.2):
    f1 = 0.5 * (1 + leak)
    f2 = 0.5 * (1 - leak)
    return f1 * X + f2 * tf.abs(X)

def bce(o, t):
    o = tf.clip_by_value(o, 1e-7, 1. - 1e-7)
    return tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(o, t))

class ReplayedGAN():
    def __init__(
            self,
            batch_size=32,
            image_shape=[24,24,1],
            dim_z=100,
            dim_y=6, #The parameters for controlling the number of events
            dim_W1=1024,
            dim_W2=128,
            dim_W3=64,
            dim_channel=1,
            num_checkpoint = 0
            ):

        self.batch_size = batch_size
        self.image_shape = image_shape
        self.dim_z = dim_z
        self.dim_y = dim_y

        self.dim_W1 = dim_W1
        self.dim_W2 = dim_W2
        self.dim_W3 = dim_W3
        self.dim_channel = dim_channel

        # Read serialized network
        graph = tf.get_default_graph()
        self.gen_W1 = graph.get_tensor_by_name("gen_W1:"+str(num_checkpoint))
        self.gen_W2 = graph.get_tensor_by_name("gen_W2:"+str(num_checkpoint))
        self.gen_W3 = graph.get_tensor_by_name("gen_W3:"+str(num_checkpoint))
        self.gen_W4 = graph.get_tensor_by_name("gen_W4:"+str(num_checkpoint))

        self.discrim_W1 = graph.get_tensor_by_name("discrim_W1:"+str(num_checkpoint))
        self.discrim_W2 = graph.get_tensor_by_name("discrim_W2:"+str(num_checkpoint))
        self.discrim_W3 = graph.get_tensor_by_name("discrim_W3:"+str(num_checkpoint))
        self.discrim_W4 = graph.get_tensor_by_name("discrim_W4:"+str(num_checkpoint))

    def samples_generator(self, batch_size):
        Z = tf.placeholder(tf.float32, [batch_size, self.dim_z])
        Y = tf.placeholder(tf.float32, [batch_size, self.dim_y])

        yb = tf.reshape(Y, [batch_size, 1, 1, self.dim_y])
        Z_ = tf.concat([Z,Y], 1)
        h1 = tf.nn.relu(batchnormalize(tf.matmul(Z_, self.gen_W1)))
        h1 = tf.concat([h1, Y], 1)
        h2 = tf.nn.relu(batchnormalize(tf.matmul(h1, self.gen_W2)))
        h2 = tf.reshape(h2, [batch_size,6,6,self.dim_W2])
        h2 = tf.concat([h2, yb*tf.ones([batch_size, 6,6, self.dim_y])], 3)

        output_shape_l3 = [batch_size,12,12,self.dim_W3]
        h3 = tf.nn.conv2d_transpose(h2, self.gen_W3, output_shape=output_shape_l3, strides=[1,2,2,1])
        h3 = tf.nn.relu( batchnormalize(h3) )
        h3 = tf.concat([h3, yb*tf.ones([batch_size, 12,12,self.dim_y])], 3)

        output_shape_l4 = [batch_size,24,24,self.dim_channel]
        h4 = tf.nn.conv2d_transpose(h3, self.gen_W4, output_shape=output_shape_l4, strides=[1,2,2,1])
        x = tf.nn.sigmoid(h4)
        return Z, Y, x


import os
import sys
import json

from .pivot_utils import chronics_to_kpi, renewableninja_to_kpi, eco2mix_to_kpi_regional, nrel_to_kpi, usa_gan_trainingset_to_kpi


def pivot_format(chronics_folder, kpi_input_folder, year, prods_charac, loads_charac, wind_solar_only, params, case):
    """
    This function contains pivot formatting of reference and synthetic chronics, in a usable format by
    kpi computation object :class: `chronix2grid.kpi.deterministic.kpis.EconomicDispatchValidator`

    Parameters
    ----------
    chronics_folder: ``str``
        path to folder which contains generated chronics
    kpi_input_folder: ``str``
        path to folder of kpi inputs, which contains paramsKPI.json and benchmark folders. paramsKPI.json tells which benchmark to read as reference
    year: ``int``
        year in which results are written
    prods_charac: :class:`pandas.DataFrame`
        characteristics of generators such as Pmax, carrier and region
    loads_charac: :class:`pandas.DataFrame`
        characteristics of loads node such as Pmax, type of demand and region
    wind_solar_only: ``bool``
        True if the generated chronics contain only wind, solar and load chronics, False otherwise
    params: ``dict``
        configuration params computed from params.json, such as timestep or mesh characteristics
    case: ``str``
        identifies the studied case for chronics generation, such as l2rpn_118

    Returns
    -------
    ref_prod: :class:`pandas.DataFrame`
        preprocessed reference productions per generator
    ref_load: :class:`pandas.DataFrame`
        preprocessed reference consumption per load node
    syn_prod: :class:`pandas.DataFrame`
        preprocessed synthetic productions per generator
    syn_load: :class:`pandas.DataFrame`
        preprocessed synthetic consumption per load node
    ref_price: :class:`pandas.DataFrame` or None
        reference price scenario (only if wind_solar_only is False)
    price: :class:`pandas.DataFrame` or None
        synthetic price scenario (only if wind_solar_only is False)
    paramsKPI: ``dict``
        dictionary with useful settings for KPI generation (e.g. season repartition, night hours)
    """

    # Read json parameters for KPI configuration
    kpi_case_input_folder = os.path.join(kpi_input_folder, case)
    json_filepath = os.path.join(kpi_case_input_folder, 'paramsKPI.json')
    with open(json_filepath, 'r') as json_file:
        paramsKPI = json.load(json_file)
    comparison = paramsKPI['comparison']
    timestep = paramsKPI['timestep']

    ## Format chosen benchmark chronics calling designed pivot functions
    if comparison == 'France':
        corresp_regions = {'R1':"Hauts-de-France", "R2": "Nouvelle-Aquitaine", "R3": "PACA"}
        if wind_solar_only:
            ref_prod, ref_load = renewableninja_to_kpi(kpi_case_input_folder, timestep, loads_charac, prods_charac, year,
                                                       params, corresp_regions, case)
        else:
            ref_prod, ref_load, ref_prices = eco2mix_to_kpi_regional(kpi_case_input_folder, timestep, prods_charac, loads_charac, year, params,
                                                         corresp_regions)
    elif comparison == 'Texas':

        if wind_solar_only:
            ref_prod, ref_load  = nrel_to_kpi(kpi_case_input_folder, timestep, prods_charac, loads_charac, params,year)
        else:
            print("Computation stopped: Texas Benchmark not implemented for a whole energy mix. Launch KPI computation in mode wind solar and load only.")
            sys.exit()
    elif comparison == 'USA - Washington State':
        if wind_solar_only:
            ref_prod, ref_load = usa_gan_trainingset_to_kpi(kpi_case_input_folder, timestep, prods_charac, loads_charac, params,year)
        else:
            print("Computation stopped: USA Benchmark not implemented for a whole energy mix. Launch KPI computation with GAN in mode wind solar and load only.")
            sys.exit()
    else:
        print("Please chose one available benchmark in paramsKPI.json/comparison. Given comparison is: "+comparison)
        sys.exit()

    # Format generated chronics
    if wind_solar_only:
        syn_prod, syn_load = chronics_to_kpi(chronics_folder, timestep, params,
                                             thermal=not wind_solar_only)
        return ref_prod, ref_load, syn_prod, syn_load, paramsKPI
    else:
        syn_prod, syn_load, prices = chronics_to_kpi(
            chronics_folder, timestep, params, thermal=not wind_solar_only)
        return ref_prod, ref_load, syn_prod, syn_load, ref_prices, prices, paramsKPI


